# copyright 2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from yams.buildobjs import EntityType, RelationDefinition, Int, ComputedRelation


class Person(EntityType):
    salary = Int()


class works_for(RelationDefinition):
    subject = "Person"
    object = "Company"
    cardinality = "?*"


class Company(EntityType):
    total_salary = Int(
        formula="Any SUM(SA) GROUPBY X WHERE " "P works_for X, P salary SA"
    )


class has_employee(ComputedRelation):
    rule = "O works_for S"
    __permissions__ = {"read": ("managers",)}



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""functional tests for core hooks

Note:
  syncschema.py hooks are mostly tested in server/test/unittest_migrations.py
"""

from datetime import datetime

from pytz import utc

from yams import ValidationError
from cubicweb import NoResultError
from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.server.hook import Operation


class CoreHooksTC(CubicWebTC):
    def test_inlined(self):
        with self.admin_access.repo_cnx() as cnx:
            self.assertEqual(self.repo.schema["sender"].inlined, True)
            cnx.execute(
                'INSERT EmailAddress X: X address "toto@logilab.fr", X alias "hop"'
            )
            cnx.execute(
                'INSERT EmailPart X: X content_format "text/plain", X ordernum 1, '
                'X content "this is a test"'
            )
            eeid = cnx.execute(
                'INSERT Email X: X messageid "<1234>", X subject "test", '
                "X sender Y, X recipients Y, X parts P "
                "WHERE Y is EmailAddress, P is EmailPart"
            )[0][0]
            cnx.execute("SET X sender Y WHERE X is Email, Y is EmailAddress")
            rset = cnx.execute(f"Any S WHERE X sender S, X eid {eeid}")
            self.assertEqual(len(rset), 1)

    def test_symmetric(self):
        with self.admin_access.repo_cnx() as cnx:
            u1 = self.create_user(cnx, "1")
            u2 = self.create_user(cnx, "2")
            u3 = self.create_user(cnx, "3")
            ga = cnx.create_entity("CWGroup", name="A")
            gb = cnx.create_entity("CWGroup", name="B")
            u1.cw_set(friend=u2)
            u2.cw_set(friend=u3)
            ga.cw_set(friend=gb)
            ga.cw_set(friend=u1)
            cnx.commit()
            for l1, l2 in (("1", "2"), ("2", "3")):
                self.assertTrue(
                    cnx.execute(
                        "Any U1,U2 WHERE U1 friend U2, U1 login %(l1)s, U2 login %(l2)s",
                        {"l1": l1, "l2": l2},
                    )
                )
                self.assertTrue(
                    cnx.execute(
                        "Any U1,U2 WHERE U2 friend U1, U1 login %(l1)s, U2 login %(l2)s",
                        {"l1": l1, "l2": l2},
                    )
                )
            self.assertTrue(
                cnx.execute('Any GA,GB WHERE GA friend GB, GA name "A", GB name "B"')
            )
            self.assertTrue(
                cnx.execute('Any GA,GB WHERE GB friend GA, GA name "A", GB name "B"')
            )
            self.assertTrue(
                cnx.execute('Any GA,U1 WHERE GA friend U1, GA name "A", U1 login "1"')
            )
            self.assertTrue(
                cnx.execute('Any GA,U1 WHERE U1 friend GA, GA name "A", U1 login "1"')
            )
            self.assertFalse(
                cnx.execute('Any GA,U WHERE GA friend U, GA name "A", U login "2"')
            )
            for l1, l2 in (("1", "3"), ("3", "1")):
                self.assertFalse(
                    cnx.execute(
                        "Any U1,U2 WHERE U1 friend U2, U1 login %(l1)s, U2 login %(l2)s",
                        {"l1": l1, "l2": l2},
                    )
                )
                self.assertFalse(
                    cnx.execute(
                        "Any U1,U2 WHERE U2 friend U1, U1 login %(l1)s, U2 login %(l2)s",
                        {"l1": l1, "l2": l2},
                    )
                )

    def test_html_tidy_hook(self):
        with self.admin_access.client_cnx() as cnx:
            entity = cnx.create_entity(
                "Workflow", name="wf1", description_format="text/html", description="yo"
            )
            self.assertEqual("yo", entity.description)
            entity = cnx.create_entity(
                "Workflow",
                name="wf2",
                description_format="text/html",
                description="<b>yo",
            )
            self.assertEqual("<b>yo</b>", entity.description)
            entity = cnx.create_entity(
                "Workflow",
                name="wf3",
                description_format="text/html",
                description="<b>yo</b>",
            )
            self.assertEqual("<b>yo</b>", entity.description)
            entity = cnx.create_entity(
                "Workflow",
                name="wf4",
                description_format="text/html",
                description="<b>R&D</b>",
            )
            self.assertEqual(
                "<b>R&amp;D</b>",
                entity.description,
            )
            entity = cnx.create_entity(
                "Workflow",
                name="wf5",
                description_format="text/html",
                description="<div>c&apos;est <b>l'ét&eacute;",
            )
            self.assertEqual("<div>c'est <b>l'été</b></div>", entity.description)

    def test_nonregr_html_tidy_hook_no_update(self):
        with self.admin_access.client_cnx() as cnx:
            entity = cnx.create_entity(
                "Workflow", name="wf1", description_format="text/html", description="yo"
            )
            entity.cw_set(name="wf2")
            self.assertEqual(entity.description, "yo")
            entity.cw_set(description="R&D<p>yo")
            self.assertEqual(entity.description, "R&amp;D<p>yo</p>")

    def test_metadata_cwuri(self):
        with self.admin_access.repo_cnx() as cnx:
            entity = cnx.create_entity("Workflow", name="wf1")
            self.assertEqual(
                entity.cwuri, self.repo.config["base-url"] + str(entity.eid)
            )

    def test_metadata_creation_modification_date(self):
        with self.admin_access.repo_cnx() as cnx:
            _now = datetime.now(utc)
            entity = cnx.create_entity("Workflow", name="wf1")
            self.assertEqual((entity.creation_date - _now).seconds, 0)
            self.assertEqual((entity.modification_date - _now).seconds, 0)

    def test_metadata_created_by(self):
        with self.admin_access.repo_cnx() as cnx:
            entity = cnx.create_entity("Bookmark", title="wf1", path="/view")
            cnx.commit()  # fire operations
            self.assertEqual(
                len(entity.created_by), 1
            )  # make sure we have only one creator
            self.assertEqual(entity.created_by[0].eid, cnx.user.eid)

    def test_metadata_owned_by(self):
        with self.admin_access.repo_cnx() as cnx:
            entity = cnx.create_entity("Bookmark", title="wf1", path="/view")
            cnx.commit()  # fire operations
            self.assertEqual(
                len(entity.owned_by), 1
            )  # make sure we have only one owner
            self.assertEqual(entity.owned_by[0].eid, cnx.user.eid)

    def test_user_login_stripped(self):
        with self.admin_access.repo_cnx() as cnx:
            u = self.create_user(cnx, "  joe  ")
            tname = cnx.execute("Any L WHERE E login L, E eid %(e)s", {"e": u.eid})[0][
                0
            ]
            self.assertEqual(tname, "joe")
            cnx.execute('SET X login " jijoe " WHERE X eid %(x)s', {"x": u.eid})
            tname = cnx.execute("Any L WHERE E login L, E eid %(e)s", {"e": u.eid})[0][
                0
            ]
            self.assertEqual(tname, "jijoe")


class AnonymousHookTC(CubicWebTC):
    anonymous_allowed = False

    def setUp(self):
        super().setUp()
        # Ensure that anonymous user is not allowed. The base test database
        # could have been created with anonymous access allowed depending on
        # the first test run.
        with self.admin_access.repo_cnx() as cnx:
            anon = cnx.find("CWUser", login="anon")
            if anon:
                anon.one().cw_delete()
                cnx.commit()

    def test_anonymous_user_is_created(self):
        with self.admin_access.repo_cnx() as cnx:
            with self.assertRaises(NoResultError):
                cnx.find("CWUser", login="anon").one()
            self.config.set_anonymous_allowed(True, anonuser="anon")
            self._init_repo()
            # This raises an error if the user does not exist
            self.assertTrue(cnx.find("CWUser", login="anon").one())


class UserGroupHooksTC(CubicWebTC):
    def test_user_group_synchronization(self):
        with self.admin_access.repo_cnx() as cnx:
            user = cnx.user
            self.assertEqual(user.groups, {"managers"})
            cnx.execute(f'SET X in_group G WHERE X eid {user.eid}, G name "guests"')
            self.assertEqual(user.groups, {"managers"})
            cnx.commit()
            self.assertEqual(user.groups, {"managers", "guests"})
            cnx.execute(f'DELETE X in_group G WHERE X eid {user.eid}, G name "guests"')
            self.assertEqual(user.groups, {"managers", "guests"})
            cnx.commit()
            self.assertEqual(user.groups, {"managers"})

    def test_user_composite_owner(self):
        with self.admin_access.repo_cnx() as cnx:
            self.create_user(cnx, "toto").eid
            # composite of euser should be owned by the euser regardless of who created it
            cnx.execute(
                'INSERT EmailAddress X: X address "toto@logilab.fr", U use_email X '
                'WHERE U login "toto"'
            )
            cnx.commit()
            self.assertEqual(
                cnx.execute(
                    "Any A WHERE X owned_by U, U use_email X,"
                    'U login "toto", X address A'
                )[0][0],
                "toto@logilab.fr",
            )

    def test_user_composite_no_owner_on_deleted_entity(self):
        with self.admin_access.repo_cnx() as cnx:
            u = self.create_user(cnx, "toto").eid
            cnx.commit()
            e = cnx.create_entity(
                "EmailAddress", address="toto@logilab.fr", reverse_use_email=u
            )
            e.cw_delete()
            cnx.commit()
            self.assertFalse(
                cnx.system_sql(
                    "SELECT * FROM owned_by_relation "
                    "WHERE eid_from NOT IN (SELECT eid FROM entities)"
                ).fetchall()
            )

    def test_no_created_by_on_deleted_entity(self):
        with self.admin_access.repo_cnx() as cnx:
            eid = cnx.execute('INSERT EmailAddress X: X address "toto@logilab.fr"')[0][
                0
            ]
            cnx.execute(f"DELETE EmailAddress X WHERE X eid {eid}")
            cnx.commit()
            self.assertFalse(
                cnx.execute("Any X WHERE X created_by Y, X eid >= %(x)s", {"x": eid})
            )


class SchemaHooksTC(CubicWebTC):
    def test_duplicate_etype_error(self):
        with self.admin_access.repo_cnx() as cnx:
            # check we can't add a CWEType or CWRType entity if it already exists one
            # with the same name
            self.assertRaises(
                ValidationError, cnx.execute, 'INSERT CWEType X: X name "CWUser"'
            )
            cnx.rollback()
            self.assertRaises(
                ValidationError, cnx.execute, 'INSERT CWRType X: X name "in_group"'
            )

    def test_validation_unique_constraint(self):
        with self.admin_access.repo_cnx() as cnx:
            with self.assertRaises(ValidationError) as cm:
                cnx.execute('INSERT CWUser X: X login "admin", X upassword "admin"')
            ex = cm.exception
            ex.translate(str)
            self.assertIsInstance(ex.entity, int)
            self.assertEqual(
                ex.errors,
                {
                    "": "some relations violate a unicity constraint",
                    "login": "login is part of violated unicity constraint",
                },
            )


class OperationTC(CubicWebTC):
    def test_bad_postcommit_event(self):
        class BadOp(Operation):
            def postcommit_event(self):
                raise RuntimeError("this is bad")

        with self.admin_access.cnx() as cnx:
            BadOp(cnx)
            with self.assertRaises(RuntimeError) as cm:
                cnx.commit()
            self.assertEqual(str(cm.exception), "this is bad")


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Hooks for synchronizing computed attributes"""


from collections import defaultdict

from rql import nodes

from cubicweb.server import hook


class RecomputeAttributeOperation(hook.DataOperationMixIn, hook.Operation):
    """Operation to recompute caches of computed attribute at commit time,
    depending on what's have been modified in the transaction and avoiding to
    recompute twice the same attribute
    """

    containercls = dict

    def add_data(self, computed_attribute, eid=None):
        try:
            self._container[computed_attribute].add(eid)
        except KeyError:
            self._container[computed_attribute] = {eid}

    def precommit_event(self):
        for computed_attribute_rdef, eids in self.get_data().items():
            attr = computed_attribute_rdef.relation_type
            formula = computed_attribute_rdef.formula
            select = self.cnx.repo.vreg.rqlhelper.parse(formula).children[0]
            xvar = select.get_variable("X")
            select.add_selected(xvar, index=0)
            select.add_group_var(xvar, index=0)
            if None in eids:
                select.add_type_restriction(xvar, computed_attribute_rdef.subject)
            else:
                select.add_eid_restriction(xvar, eids)
            update_rql = "SET X %s %%(value)s WHERE X eid %%(x)s" % attr
            for eid, value in self.cnx.execute(select.as_string()):
                self.cnx.execute(update_rql, {"value": value, "x": eid})


class EntityWithCACreatedHook(hook.Hook):
    """When creating an entity that has some computed attribute, those
    attributes have to be computed.

    Concret class of this hook are generated at registration time by
    introspecting the schema.
    """

    __abstract__ = True
    events = ("after_add_entity",)
    # list of computed attribute rdefs that have to be recomputed
    computed_attributes = None

    def __call__(self):
        for rdef in self.computed_attributes:
            RecomputeAttributeOperation.get_instance(self._cw).add_data(
                rdef, self.entity.eid
            )


class RelationInvolvedInCAModifiedHook(hook.Hook):
    """When some relation used in a computed attribute is updated, those
    attributes have to be recomputed.

    Concret class of this hook are generated at registration time by
    introspecting the schema.
    """

    __abstract__ = True
    events = ("after_add_relation", "before_delete_relation")
    # list of (computed attribute rdef, optimize_on) that have to be recomputed
    optimized_computed_attributes = None

    def __call__(self):
        for rdef, optimize_on in self.optimized_computed_attributes:
            if optimize_on is None:
                eid = None
            else:
                eid = getattr(self, optimize_on)
            RecomputeAttributeOperation.get_instance(self._cw).add_data(rdef, eid)


class AttributeInvolvedInCAModifiedHook(hook.Hook):
    """When some attribute used in a computed attribute is updated, those
    attributes have to be recomputed.

    Concret class of this hook are generated at registration time by
    introspecting the schema.
    """

    __abstract__ = True
    events = ("after_update_entity",)
    # list of (computed attribute rdef, attributes of this entity type involved)
    # that may have to be recomputed
    attributes_computed_attributes = None

    def __call__(self):
        edited_attributes = frozenset(self.entity.cw_edited)
        for rdef, used_attributes in self.attributes_computed_attributes.items():
            if edited_attributes.intersection(used_attributes):
                # XXX optimize if the modified attributes belong to the same
                # entity as the computed attribute
                RecomputeAttributeOperation.get_instance(self._cw).add_data(rdef)


# code generation at registration time #########################################


def _optimize_on(formula_select, rtype):
    """Given a formula and some rtype, tells whether on update of the given
    relation, formula may be recomputed only for rhe relation's subject
    ('eidfrom' returned), object ('eidto' returned) or None.

    Optimizing is only possible when X is used as direct subject/object of this
    relation, else we may miss some necessary update.
    """
    for rel in formula_select.get_nodes(nodes.Relation):
        if rel.r_type == rtype:
            sub = rel.get_variable_parts()[0]
            obj = rel.get_variable_parts()[1]
            if sub.name == "X":
                return "eidfrom"
            elif obj.name == "X":
                return "eidto"
            else:
                return None


class _FormulaDependenciesMatrix:
    """This class computes and represents the dependencies of computed attributes
    towards relations and attributes
    """

    def __init__(self, schema):
        """Analyzes the schema to compute the dependencies"""
        # entity types holding some computed attribute {etype: [computed rdefs]}
        self.computed_attribute_by_etype = defaultdict(list)
        # depending entity types {dep. etype: {computed rdef: dep. etype attributes}}
        self.computed_attribute_by_etype_attrs = defaultdict(lambda: defaultdict(set))
        # depending relations def {dep. rdef: [computed rdefs]
        self.computed_attribute_by_relation = defaultdict(list)  # by rdef
        # Walk through all attributes definitions
        for rdef in schema.iter_computed_attributes():
            self.computed_attribute_by_etype[rdef.subject.type].append(rdef)
            # extract the relations it depends upon - `rdef.formula_select` is
            # expected to have been set by finalize_computed_attributes
            select = rdef.formula_select
            for rel_node in select.get_nodes(nodes.Relation):
                if rel_node.is_types_restriction():
                    continue
                rschema = schema.relation_schema_for(rel_node.r_type)
                lhs, rhs = rel_node.get_variable_parts()
                for sol in select.solutions:
                    subject_etype = sol[lhs.name]
                    if isinstance(rhs, nodes.VariableRef):
                        object_etypes = {sol[rhs.name] for sol in select.solutions}
                    else:
                        object_etypes = rschema.objects(subject_etype)
                    for object_etype in object_etypes:
                        if rschema.final:
                            attr_for_computations = (
                                self.computed_attribute_by_etype_attrs[subject_etype]
                            )
                            attr_for_computations[rdef].add(rschema.type)
                        else:
                            depend_on_rdef = rschema.relation_definitions[
                                subject_etype, object_etype
                            ]
                            self.computed_attribute_by_relation[depend_on_rdef].append(
                                rdef
                            )

    def generate_entity_creation_hooks(self):
        for etype, computed_attributes in self.computed_attribute_by_etype.items():
            regid = f"computed_attribute.{etype}_created"
            selector = hook.is_instance(etype)
            yield type(
                f"{etype}CreatedHook",
                (EntityWithCACreatedHook,),
                {
                    "__regid__": regid,
                    "__select__": hook.Hook.__select__ & selector,
                    "computed_attributes": computed_attributes,
                },
            )

    def generate_relation_change_hooks(self):
        for rdef, computed_attributes in self.computed_attribute_by_relation.items():
            regid = f"computed_attribute.{rdef.relation_type}_modified"
            selector = hook.match_rtype(
                rdef.relation_type.type,
                frometypes=(rdef.subject.type,),
                toetypes=(rdef.object.type,),
            )
            optimized_computed_attributes = []
            for computed_rdef in computed_attributes:
                optimized_computed_attributes.append(
                    (
                        computed_rdef,
                        _optimize_on(computed_rdef.formula_select, rdef.relation_type),
                    )
                )
            yield type(
                f"{rdef.relation_type}ModifiedHook",
                (RelationInvolvedInCAModifiedHook,),
                {
                    "__regid__": regid,
                    "__select__": hook.Hook.__select__ & selector,
                    "optimized_computed_attributes": optimized_computed_attributes,
                },
            )

    def generate_entity_update_hooks(self):
        for (
            etype,
            attributes_computed_attributes,
        ) in self.computed_attribute_by_etype_attrs.items():
            regid = f"computed_attribute.{etype}_updated"
            selector = hook.is_instance(etype)
            yield type(
                f"{etype}ModifiedHook",
                (AttributeInvolvedInCAModifiedHook,),
                {
                    "__regid__": regid,
                    "__select__": hook.Hook.__select__ & selector,
                    "attributes_computed_attributes": attributes_computed_attributes,
                },
            )


def registration_callback(vreg):
    vreg.register_all(globals().values(), __name__)
    dependencies = _FormulaDependenciesMatrix(vreg.schema)
    for hook_class in dependencies.generate_entity_creation_hooks():
        vreg.register(hook_class)
    for hook_class in dependencies.generate_relation_change_hooks():
        vreg.register(hook_class)
    for hook_class in dependencies.generate_entity_update_hooks():
        vreg.register(hook_class)



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""cubicweb.migration unit tests"""

from os.path import dirname, join
from unittest.mock import patch

from logilab.common.shellutils import ASK
from logilab.common.testlib import TestCase, unittest_main

from cubicweb import utils
from cubicweb.cwconfig import CubicWebConfiguration
from cubicweb.devtools import get_test_db_handler
from cubicweb.devtools.apptest_config import (
    ApptestConfiguration,
)
from cubicweb.migration import (
    filter_scripts,
    split_constraint,
    version_strictly_lower,
    MigrationHelper,
)


class Schema(dict):
    def has_entity(self, e_type):
        return e_type in self


SMIGRDIR = join(dirname(__file__), "data", "server_migration") + "/"
TMIGRDIR = join(dirname(__file__), "data", "migration") + "/"


class MigrTestConfig(ApptestConfiguration):
    verbosity = 0

    def migration_scripts_dir(cls):
        return SMIGRDIR

    def cube_migration_scripts_dir(cls, cube):
        return TMIGRDIR


class MigrationToolsTC(TestCase):
    def setUp(self):
        self.config = MigrTestConfig("data", __file__)
        from yams.schema import Schema

        self.config.load_schema = lambda expand_cubes=False: Schema("test")
        self.config.__class__.cubicweb_appobject_path = frozenset()
        self.config.__class__.cube_appobject_path = frozenset()
        self.config.__class__.name = "all-in-one"

    def test_filter_scripts_base(self):
        self.assertListEqual(
            filter_scripts(self.config, SMIGRDIR, (2, 3, 0), (2, 4, 0)), []
        )
        self.assertListEqual(
            filter_scripts(self.config, SMIGRDIR, (2, 4, 0), (2, 5, 0)),
            [((2, 5, 0), SMIGRDIR + "2.5.0_Any.sql")],
        )
        self.assertListEqual(
            filter_scripts(self.config, SMIGRDIR, (2, 5, 0), (2, 6, 0)),
            [((2, 6, 0), SMIGRDIR + "2.6.0_Any.sql")],
        )
        self.assertListEqual(
            filter_scripts(self.config, SMIGRDIR, (2, 4, 0), (2, 6, 0)),
            [
                ((2, 5, 0), SMIGRDIR + "2.5.0_Any.sql"),
                ((2, 6, 0), SMIGRDIR + "2.6.0_Any.sql"),
            ],
        )
        self.assertListEqual(
            filter_scripts(self.config, SMIGRDIR, (2, 5, 0), (2, 5, 1)), []
        )
        self.assertListEqual(
            filter_scripts(self.config, SMIGRDIR, (2, 5, 0), (2, 10, 2)),
            [
                ((2, 6, 0), SMIGRDIR + "2.6.0_Any.sql"),
                ((2, 10, 2), SMIGRDIR + "2.10.2_Any.sql"),
            ],
        )
        self.assertListEqual(
            filter_scripts(self.config, SMIGRDIR, (2, 5, 1), (2, 6, 0)),
            [((2, 6, 0), SMIGRDIR + "2.6.0_Any.sql")],
        )

        self.assertListEqual(
            filter_scripts(self.config, TMIGRDIR, (0, 0, 2), (0, 0, 3)),
            [((0, 0, 3), TMIGRDIR + "0.0.3_Any.py")],
        )
        self.assertListEqual(
            filter_scripts(self.config, TMIGRDIR, (0, 0, 2), (0, 0, 4)),
            [
                ((0, 0, 3), TMIGRDIR + "0.0.3_Any.py"),
                ((0, 0, 4), TMIGRDIR + "0.0.4_Any.py"),
            ],
        )

    def test_filter_scripts_for_mode(self):
        config = CubicWebConfiguration("data", __file__)
        config.verbosity = 0
        config = self.config
        self.assertListEqual(
            filter_scripts(config, TMIGRDIR, (0, 0, 4), (0, 1, 0)),
            [
                ((0, 1, 0), TMIGRDIR + "0.1.0_Any.py"),
                ((0, 1, 0), TMIGRDIR + "0.1.0_common.py"),
                ((0, 1, 0), TMIGRDIR + "0.1.0_repository.py"),
            ],
        )

    def test_version_strictly_lower(self):
        self.assertTrue(version_strictly_lower(None, "1.0.0"))
        self.assertFalse(version_strictly_lower("1.0.0", None))


class BaseCreationTC(TestCase):
    def test_db_creation(self):
        """make sure database can be created"""
        config = ApptestConfiguration("data", __file__)
        source = config.system_source_config
        self.assertEqual(source["db-driver"], "sqlite")
        handler = get_test_db_handler(config)
        handler.init_test_database()
        handler.build_db_cache()
        repo, cnx = handler.get_repo_and_cnx()
        with cnx:
            self.assertEqual(
                cnx.execute(
                    'Any SN WHERE X is CWUser, X login "admin", X in_state S, S name SN'
                ).rows,
                [["activated"]],
            )
        repo.shutdown()


def test_split_constraint():
    assert split_constraint(">=0.1.0") == (">=", "0.1.0")
    assert split_constraint(">= 0.1.0") == (">=", "0.1.0")
    assert split_constraint(">0.1.1") == (">", "0.1.1")
    assert split_constraint("> 0.1.1") == (">", "0.1.1")
    assert split_constraint("<0.2.0") == ("<", "0.2.0")
    assert split_constraint("< 0.2.0") == ("<", "0.2.0")
    assert split_constraint("<=42.1.0") == ("<=", "42.1.0")
    assert split_constraint("<= 42.1.0") == ("<=", "42.1.0")


class WontColideWithOtherExceptionsException(Exception):
    pass


class MigrationHelperTC(TestCase):
    @patch.object(utils, "get_pdb")
    @patch.object(ASK, "ask", return_value="pdb")
    def test_confirm_no_traceback(self, ask, get_pdb):
        post_mortem = get_pdb.return_value.post_mortem
        set_trace = get_pdb.return_value.set_trace

        # we need to break after post_mortem is called otherwise we get
        # infinite recursion
        set_trace.side_effect = WontColideWithOtherExceptionsException

        mh = MigrationHelper(config=None)

        with self.assertRaises(WontColideWithOtherExceptionsException):
            mh.confirm("some question")

        get_pdb.assert_called_once()
        set_trace.assert_called_once()
        post_mortem.assert_not_called()

    @patch.object(utils, "get_pdb")
    @patch.object(ASK, "ask", return_value="pdb")
    def test_confirm_got_traceback(self, ask, get_pdb):
        post_mortem = get_pdb.return_value.post_mortem
        set_trace = get_pdb.return_value.set_trace

        # we need to break after post_mortem is called otherwise we get
        # infinite recursion
        post_mortem.side_effect = WontColideWithOtherExceptionsException

        mh = MigrationHelper(config=None)

        fake_traceback = object()

        with self.assertRaises(WontColideWithOtherExceptionsException):
            mh.confirm("some question", traceback=fake_traceback)

        get_pdb.assert_called_once()
        set_trace.assert_not_called()
        post_mortem.assert_called_once()

        # we want post_mortem to actually receive the traceback
        self.assertEqual(post_mortem.call_args, ((fake_traceback,),))


if __name__ == "__main__":
    unittest_main()



# copyright 2019 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
#
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

"""
Debug view for pyramid debug toolbar and others to help development
"""

import inspect
import logging
import os
from itertools import dropwhile

from mako.template import Template
from pyramid.response import Response

from cubicweb.misc.source_highlight import highlight_html, generate_css, has_pygments

DEBUG_DISPLAY_SOURCE_CODE_PATH = "_debug_display_source_code"

FILES_WHITE_LIST = set()


def _generate_link_to_source(file_path, start=None, end=None, tag_body="[source]"):
    if start:
        # step back a bit so we have a bit of top padding wen displaying the page
        # and the highlighted line isn't glued to top of the browser window
        line_anchor = max(0, start - 10)

        if end:
            return (
                '<a href="../%s?file=%s&line=%s&end=%s#line-%s" target="_blank">%s</a>'
                % (
                    DEBUG_DISPLAY_SOURCE_CODE_PATH,
                    file_path,
                    start,
                    end,
                    line_anchor,
                    tag_body,
                )
            )
        else:
            return (
                '<a href="../{}?file={}&line={}#line-{}" target="_blank">{}</a>'.format(
                    DEBUG_DISPLAY_SOURCE_CODE_PATH,
                    file_path,
                    start,
                    line_anchor,
                    tag_body,
                )
            )

    return '<a href="../{}?file={}" target="_blank">{}</a>'.format(
        DEBUG_DISPLAY_SOURCE_CODE_PATH,
        file_path,
        tag_body,
    )


def source_code_url(object_or_class):
    if object_or_class is None:
        return ""

    if not inspect.isclass(object_or_class):
        object_or_class = object_or_class.__class__

    try:
        file_path = inspect.getsourcefile(object_or_class)
    except TypeError:
        logging.debug(f"Error while trying to source code of '{object_or_class}'")
        return ""

    FILES_WHITE_LIST.add(file_path)

    try:
        source_code, line = inspect.getsourcelines(object_or_class)
    except OSError:  # when we couldn't read the source code/line
        return _generate_link_to_source(file_path)

    return _generate_link_to_source(file_path, line, line + len(source_code))


def source_code_url_in_stack(stack, highlighted=False):
    new_stack = []

    if highlighted:
        for i in stack.split(" File "):
            # expecting this format:
            # '<span class="nb">&quot;/path/to/file.py&quot;</span>,
            # line <span class="m">885</span>,...'
            if not i.startswith('<span class="nb">'):
                new_stack.append(i)
                continue

            # this will give:
            # ['<span class="nb">', '/file/to/path.py', '</span>, ...']
            tag, file_path, rest = i.split("&quot;", 2)

            # "rest" is like that: '</span>, line <span class="m">885</span>, ...'
            # we want to extrait "885" here
            line_number = int(
                "".join(dropwhile(lambda x: not x.isdigit(), rest)).split("<")[0]
            )

            new_stack.append(
                "%s%s%s"
                % (
                    tag,
                    _generate_link_to_source(
                        file_path,
                        start=line_number,
                        tag_body="&quot;%s&quot;" % file_path,
                    ),
                    rest,
                )
            )

            FILES_WHITE_LIST.add(file_path)

        new_stack = " File ".join(new_stack)

    # no syntax
    else:
        for i in stack.split("\n"):
            # expecting this format:
            # File "/path/to/file.py", line 885, in stuf\n  some_code\nFile "/stuff.py", line...
            if not i.startswith("  File "):
                new_stack.append(i)
                continue

            # this will give:
            # ['File "', '/path/to/file.py', '", line 885, in stuf']
            beginning, file_path, rest = i.split('"', 2)
            line_number = int(
                "".join(dropwhile(lambda x: not x.isdigit(), rest)).split(",")[0]
            )

            new_stack.append(
                "%s%s%s"
                % (
                    beginning,
                    _generate_link_to_source(
                        file_path, start=line_number, tag_body='"%s"' % file_path
                    ),
                    rest,
                )
            )

            FILES_WHITE_LIST.add(file_path)

        new_stack = "\n".join(new_stack)

    return new_stack


def debug_display_source_code(request):
    """
    This view display a python source file content for making debugging easier.

    It is only activated on debug mode (-D) for pyramid during development.

    It will uses pygment if installed to colorize the source code.
    """

    if "file" not in request.params:
        return Response(
            "Error: you should have end up on this page following a link in the "
            "pyramid debug toolbar with the correct parameters."
        )

    source_code_file = request.params["file"]

    if not os.path.exists(source_code_file):
        return Response(
            f"Error: file '{source_code_file}' doesn't exist on the filesystem."
        )

    # security
    if source_code_file not in FILES_WHITE_LIST:
        return Response("Error: access to file is not authorized")

    try:
        content = open(source_code_file).read()
    except Exception as e:
        return Response(
            f"Error: while opening file '{source_code_file}' got the error: {e}"
        )

    lines = []
    line_begin = request.params.get("line", None)
    if line_begin:
        line_end = request.params.get("end", line_begin)
        lines = list(range(int(line_begin), int(line_end) + 1))

    this_file_directory = os.path.split(os.path.realpath(__file__))[0]
    template_filename = os.path.join(
        this_file_directory, "debug_toolbar_templates/debug_source_code.mako"
    )

    html = Template(filename=template_filename).render(
        file_path=source_code_file,
        content=content,
        has_pygments=has_pygments,
        highlight_html=highlight_html,
        css=generate_css(),
        lines=lines,
    )

    return Response(html)



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Some utilities for CubicWeb server/clients."""

import base64
import datetime
import decimal
import json
import random
import re
from functools import wraps
from inspect import getfullargspec as getargspec
from itertools import repeat
from logging import getLogger
from uuid import uuid4

from logilab.common.date import ustrftime
from logilab.mtconverter import xml_escape

from cubicweb import Binary

_MARKER = object()

# initialize random seed from current time
random.seed()


def format_and_escape_string(value, *args, escape=True):
    if args:
        try:
            if escape:
                if isinstance(args[0], dict):
                    value = value % {
                        key: xml_escape(str(value)) if value else value
                        for key, value in args[0].items()
                    }
                else:
                    value = value % tuple(xml_escape(str(x)) if x else x for x in args)
            else:
                if isinstance(args[0], dict):
                    value = value % args[0]
                else:
                    value = value % args
        except Exception:
            # for debugging, the exception is often hidden otherwise
            import traceback

            traceback.print_exc()
            print(f'text: "{value}"')
            print(f'arguments: "{args}"')
            raise

    return value


def admincnx(appid):
    from cubicweb import repoapi
    from cubicweb.cwconfig import CubicWebConfiguration
    from cubicweb.server.repository import Repository

    config = CubicWebConfiguration.config_for(appid)

    login = config.default_admin_config["login"]
    password = config.default_admin_config["password"]

    repo = Repository(config)
    repo.bootstrap()
    return repoapi.connect(repo, login, password=password)


def make_uid(key=None):
    """Return a unique identifier string.

    if specified, `key` is used to prefix the generated uid so it can be used
    for instance as a DOM id or as sql table name.

    See uuid.uuid4 documentation for the shape of the generated identifier, but
    this is basically a 32 bits hexadecimal string.
    """
    if key is None:
        return uuid4().hex
    return str(key) + uuid4().hex


def support_args(callable, *argnames):
    """return true if the callable support given argument names"""
    if isinstance(callable, type):
        callable = callable.__init__
    argspec = getargspec(callable)
    if argspec[2]:
        return True
    for argname in argnames:
        if argname not in argspec[0]:
            return False
    return True


class wrap_on_write:
    """Sometimes it is convenient to NOT write some container element
    if it happens that there is nothing to be written within,
    but this cannot be known beforehand.
    Hence one can do this:

    .. sourcecode:: python

       with wrap_on_write(w, '<div class="foo">', '</div>') as wow:
           component.render_stuff(wow)
    """

    def __init__(self, w, tag, closetag=None):
        self.written = False
        self.tag = tag
        self.closetag = closetag
        self.w = w

    def __enter__(self):
        return self

    def __call__(self, data, *args, escape=True):
        if self.written is False:
            self.w(self.tag)
            self.written = True
        self.w(data, *args, escape=escape)

    def __exit__(self, exctype, value, traceback):
        if self.written is True:
            if self.closetag:
                self.w(self.closetag)
            else:
                self.w(self.tag.replace("<", "</", 1))


# use networkX instead ?
# http://networkx.lanl.gov/reference/algorithms.traversal.html#module-networkx.algorithms.traversal.astar
def transitive_closure_of(entity, rtype, _seen=None):
    """return transitive closure *for the subgraph starting from the given
    entity* (eg 'parent' entities are not included in the results)
    """
    if _seen is None:
        _seen = set()
    _seen.add(entity.eid)
    yield entity
    for child in getattr(entity, rtype):
        if child.eid in _seen:
            continue
        for subchild in transitive_closure_of(child, rtype, _seen):
            yield subchild


class RepeatList:
    """fake a list with the same element in each row"""

    __slots__ = ("_size", "_item")

    def __init__(self, size, item):
        self._size = size
        self._item = item

    def __repr__(self):
        return "<cubicweb.utils.RepeatList at %s item=%s size=%s>" % (
            id(self),
            self._item,
            self._size,
        )

    def __len__(self):
        return self._size

    def __iter__(self):
        return repeat(self._item, self._size)

    def __getitem__(self, index):
        if isinstance(index, slice):
            # XXX could be more efficient, but do we bother?
            return ([self._item] * self._size)[index]
        return self._item

    def __delitem__(self, idc):
        assert self._size > 0
        self._size -= 1

    def __add__(self, other):
        if isinstance(other, RepeatList):
            if other._item == self._item:
                return RepeatList(self._size + other._size, self._item)
            return ([self._item] * self._size) + other[:]
        return ([self._item] * self._size) + other

    def __radd__(self, other):
        if isinstance(other, RepeatList):
            if other._item == self._item:
                return RepeatList(self._size + other._size, self._item)
            return other[:] + ([self._item] * self._size)
        return other[:] + ([self._item] * self._size)

    def __eq__(self, other):
        if isinstance(other, RepeatList):
            return other._size == self._size and other._item == self._item
        return self[:] == other

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        raise NotImplementedError

    def pop(self, i):
        self._size -= 1


def handle_writing_constraints(method):
    @wraps(method)
    def wrapper(self, value, *args, escape=True):
        if self.tracewrites:
            from traceback import format_stack

            stack = format_stack(None)[:-1]
            escaped_stack = xml_escape(json_dumps("\n".join(stack)))
            escaped_html = xml_escape(value).replace("\n", "<br/>\n")
            tpl = '<span onclick="alert(%s)">%s</span>'
            value = tpl % (escaped_stack, escaped_html)
        return method(self, value, *args, escape=escape)

    return wrapper


class UStringIO(list):
    """a file wrapper which automatically encode unicode string to an encoding
    specifed in the constructor
    """

    def __init__(self, tracewrites=False, *args, **kwargs):
        self.tracewrites = tracewrites
        super(UStringIO, self).__init__(*args, **kwargs)

    def __bool__(self):
        return True

    __nonzero__ = __bool__

    def write(self, value, *args, escape=True):
        if self.tracewrites:
            from traceback import format_stack

            stack = format_stack(None)[:-1]
            escaped_stack = xml_escape(json_dumps("\n".join(stack)))
            escaped_html = xml_escape(value).replace("\n", "<br/>\n")
            tpl = '<span onclick="alert(%s)">%s</span>'
            value = tpl % (escaped_stack, escaped_html)

        value = format_and_escape_string(value, *args, escape=escape)

        self.append(value)

    @handle_writing_constraints
    def write_front(self, value, *args, escape=True):
        value = format_and_escape_string(value, *args, escape=escape)

        self.insert(0, value)

    def getvalue(self):
        return "".join(self)

    def __repr__(self):
        return "<%s at %#x>" % (self.__class__.__name__, id(self))


class CubicWebJsonEncoder(json.JSONEncoder):
    """define a json encoder to be able to encode yams std types"""

    def default(self, obj):
        if hasattr(obj, "__json_encode__"):
            return xml_escape_from_dict(obj.__json_encode__())
        if isinstance(obj, datetime.datetime):
            return ustrftime(obj, "%Y/%m/%d %H:%M:%S")
        elif isinstance(obj, datetime.date):
            return ustrftime(obj, "%Y/%m/%d")
        elif isinstance(obj, datetime.time):
            return obj.strftime("%H:%M:%S")
        elif isinstance(obj, datetime.timedelta):
            return (obj.days * 24 * 60 * 60) + obj.seconds
        elif isinstance(obj, decimal.Decimal):
            return float(obj)
        elif isinstance(obj, Binary):
            return base64.b64encode(obj.getvalue()).decode("ascii")
        try:
            return json.JSONEncoder.default(self, obj)
        except TypeError:
            # we never ever want to fail because of an unknown type,
            # just return None in those cases.
            return None


def json_dumps(value, **kwargs):
    return json.dumps(value, cls=CubicWebJsonEncoder, **kwargs)


def xml_escape_from_dict(data):
    for key, value in data.items():
        if isinstance(value, str):
            data[key] = xml_escape(value)
    return data


class JSString(str):
    """use this string sub class in values given to :func:`js_dumps` to
    insert raw javascript chain in some JSON string
    """


def _dict2js(d, predictable=False):
    if predictable:
        it = sorted(d.items())
    else:
        it = d.items()
    res = [
        js_dumps(key, predictable) + ": " + js_dumps(val, predictable)
        for key, val in it
    ]
    return "{%s}" % ", ".join(res)


def _list2js(a_list, predictable=False):
    return f"[{', '.join([js_dumps(val, predictable) for val in a_list])}]"


def js_dumps(something, predictable=False):
    """similar as :func:`json_dumps`, except values which are instances of
    :class:`JSString` are expected to be valid javascript and will be output
    as is

    >>> js_dumps({'hop': JSString('$.hop'), 'bar': None}, predictable=True)
    '{bar: null, hop: $.hop}'
    >>> js_dumps({'hop': '$.hop'})
    '{hop: "$.hop"}'
    >>> js_dumps({'hip': {'hop': JSString('momo')}})
    '{hip: {hop: momo}}'
    """
    if isinstance(something, dict):
        return _dict2js(something, predictable)
    if isinstance(something, list):
        return _list2js(something, predictable)
    if isinstance(something, JSString):
        return something
    return json_dumps(something, sort_keys=predictable)


PERCENT_IN_URLQUOTE_RE = re.compile(r"%(?=[0-9a-fA-F]{2})")


def js_href(javascript_code):
    """Generate a "javascript: ..." string for an href attribute.

    Some % which may be interpreted in a href context will be escaped.

    In an href attribute, url-quotes-looking fragments are interpreted before
    being given to the javascript engine. Valid url quotes are in the form
    ``%xx`` with xx being a byte in hexadecimal form. This means that ``%toto``
    will be unaltered but ``%babar`` will be mangled because ``ba`` is the
    hexadecimal representation of 186.

    >>> js_href('alert("babar");')
    'javascript: alert("babar");'
    >>> js_href('alert("%babar");')
    'javascript: alert("%25babar");'
    >>> js_href('alert("%toto %babar");')
    'javascript: alert("%toto %25babar");'
    >>> js_href('alert("%1337%");')
    'javascript: alert("%251337%");'
    """
    return "javascript: " + PERCENT_IN_URLQUOTE_RE.sub(r"%25", javascript_code)


def get_pdb():
    "return ipdb if its installed, otherwise pdb"
    try:
        import ipdb
    except ImportError:
        import pdb

        return pdb
    else:
        return ipdb


logger = getLogger("cubicweb.utils")



if confirm("use Europe/Paris as timezone?"):
    timezone = "Europe/Paris"
else:
    import pytz

    while True:
        timezone = input("enter your timezone")
        if timezone in pytz.common_timezones:
            break

dbdriver = repo.system_source.dbdriver
if dbdriver == "postgres":
    sql(f"SET TIME ZONE '{timezone}'")

for entity in schema.entities():
    if entity.final or entity.type not in fsschema:
        continue
    change_attribute_type(entity.type, "creation_date", "TZDatetime", ask_confirm=False)
    change_attribute_type(
        entity.type, "modification_date", "TZDatetime", ask_confirm=False
    )

if dbdriver == "postgres":
    sql("SET TIME ZONE UTC")



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
from yams.buildobjs import (
    EntityType,
    RelationDefinition,
    String,
    SubjectRelation,
    ComputedRelation,
    Int,
)

from cubicweb.schema import ERQLExpression


class Person(EntityType):
    name = String()


class Affaire(EntityType):
    __permissions__ = {
        "read": (
            "managers",
            ERQLExpression("X owned_by U"),
            ERQLExpression("X concerne S?, S owned_by U"),
        ),
        "add": ("managers", ERQLExpression("X concerne S, S owned_by U")),
        "update": (
            "managers",
            "owners",
            ERQLExpression('X in_state S, S name in ("pitetre", "en cours")'),
        ),
        "delete": ("managers", "owners", ERQLExpression("X concerne S, S owned_by U")),
    }
    ref = String(fulltextindexed=True, indexed=True, maxsize=16)
    documented_by = SubjectRelation("Card", cardinality="1*")
    concerne = SubjectRelation(("Societe", "Note"), cardinality="1*")


class Societe(EntityType):
    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "update": ("managers", "owners", ERQLExpression("U login L, X nom L")),
        "delete": ("managers", "owners", ERQLExpression("U login L, X nom L")),
        "add": (
            "managers",
            "users",
        ),
    }
    nom = String()


class Division(Societe):
    __specializes_schema__ = True


class Note(EntityType):
    pass


class require_permission(RelationDefinition):
    subject = ("Card", "Note")
    object = "CWPermission"


class require_state(RelationDefinition):
    subject = "CWPermission"
    object = "State"


class inlined_card(RelationDefinition):
    subject = "Affaire"
    object = "Card"
    inlined = True
    cardinality = "?*"


class inlined_note(RelationDefinition):
    subject = "Card"
    object = "Note"
    inlined = True
    cardinality = "?*"


class inlined_affaire(RelationDefinition):
    subject = "Note"
    object = "Affaire"
    inlined = True
    cardinality = "?*"


class responsable(RelationDefinition):
    subject = "Societe"
    object = "CWUser"
    inlined = True
    cardinality = "1*"


class Contribution(EntityType):
    code = Int()


class ArtWork(EntityType):
    name = String()


class Role(EntityType):
    name = String()


class contributor(RelationDefinition):
    subject = "Contribution"
    object = "Person"
    cardinality = "1*"
    inlined = True


class manifestation(RelationDefinition):
    subject = "Contribution"
    object = "ArtWork"


class role(RelationDefinition):
    subject = "Contribution"
    object = "Role"


class illustrator_of(ComputedRelation):
    rule = (
        "C is Contribution, C contributor S, C manifestation O, "
        'C role R, R name "illustrator"'
    )


class participated_in(ComputedRelation):
    rule = "S contributor O"


class match(RelationDefinition):
    subject = "ArtWork"
    object = "Note"



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Always executed before all others in server migration

it should only include low level schema changes
"""
from cubicweb import ConfigurationError
from cubicweb.server import schemaserial as ss

applcubicwebversion, cubicwebversion = versions_map["cubicweb"]


def _add_relation_definition_no_perms(subjtype, rtype, objtype):
    rschema = fsschema.relation_schema_for(rtype)
    rdef = rschema.relation_definitions[(subjtype, objtype)]
    rdef.rtype = schema.relation_schema_for(rtype)
    rdef.subject = schema.entity_schema_for(subjtype)
    rdef.object = schema.entity_schema_for(objtype)
    ss.execschemarql(rql, rdef, ss.rdef2rql(rdef, CSTRMAP, groupmap=None))
    commit(ask_confirm=False)


def replace_eid_sequence_with_eid_numrange(session):
    dbh = session.repo.system_source.dbhelper
    cursor = session.cnxset.cu
    try:
        cursor.execute(dbh.sql_sequence_current_state("entities_id_seq"))
        lasteid = cursor.fetchone()[0]
    except Exception:  # programming error, already migrated
        session.rollback()
        return

    cursor.execute(dbh.sql_drop_sequence("entities_id_seq"))
    cursor.execute(dbh.sql_create_numrange("entities_id_seq"))
    cursor.execute(dbh.sql_restart_numrange("entities_id_seq", initial_value=lasteid))
    session.commit()


if applcubicwebversion < (3, 19, 0) and cubicwebversion >= (3, 19, 0):
    try:
        # need explicit drop of the indexes on some database systems (sqlserver)
        sql(repo.system_source.dbhelper.sql_drop_index("entities", "mtime"))
        sql('ALTER TABLE "entities" DROP COLUMN "mtime"')
        sql('ALTER TABLE "entities" DROP COLUMN "source"')
    except Exception:  # programming error, already migrated
        print("Failed to drop mtime or source database columns")
        print("'entities' table of the database has probably been already updated")

    commit()

    replace_eid_sequence_with_eid_numrange(session)


if applcubicwebversion < (3, 18, 0) and cubicwebversion >= (3, 18, 0):
    driver = config.system_source_config["db-driver"]
    if not driver == "postgres":
        import sys

        print(
            "This migration is not supported for backends other than postgres (yet).",
            file=sys.stderr,
        )
        sys.exit(1)

    add_relation_definition("CWAttribute", "add_permission", "CWGroup")
    add_relation_definition("CWAttribute", "add_permission", "RQLExpression")

    # a bad defaultval in 3.13.8 schema was fixed in 3.13.9, but the migration was missed
    rql(
        'SET ATTR defaultval NULL WHERE ATTR from_entity E, E name "CWSource", ATTR relation_type T, T name "in_synchronization"'
    )

    # the migration gets confused when we change rdefs out from under it.  So
    # explicitly remove this size constraint so it doesn't stick around and break
    # things later.
    rdefeid = schema["defaultval"].relation_definitions.values()[0].eid
    rql(
        'DELETE CWConstraint C WHERE C cstrtype T, T name "SizeConstraint", R constrained_by C, R eid %(eid)s',
        {"eid": rdefeid},
    )

    sync_schema_props_perms("defaultval")

    def convert_defaultval(cwattr, default):
        from decimal import Decimal
        import yams
        from cubicweb import Binary

        if default is None:
            return
        if isinstance(default, Binary):
            # partially migrated instance, try to be idempotent
            return default
        atype = cwattr.to_entity[0].name
        if atype == "Boolean":
            # boolean attributes with default=False were stored as ''
            assert default in ("True", "False", ""), repr(default)
            default = default == "True"
        elif atype in ("Int", "BigInt"):
            default = int(default)
        elif atype == "Float":
            default = float(default)
        elif atype == "Decimal":
            default = Decimal(default)
        elif atype in ("Date", "Datetime", "TZDatetime", "Time"):
            try:
                # handle NOW and TODAY, keep them stored as strings
                yams.KEYWORD_MAP[atype][default.upper()]
                default = default.upper()
            except KeyError:
                # otherwise get an actual date or datetime
                default = yams.DATE_FACTORY_MAP[atype](default)
        else:
            assert atype == "String", atype
        return Binary.zpickle(default)

    dbh = repo.system_source.dbhelper

    sql(f"ALTER TABLE cw_cwattribute ADD new_defaultval {dbh.TYPE_MAPPING['Bytes']}")

    for cwattr in rql("CWAttribute X").entities():
        olddefault = cwattr.defaultval
        if olddefault is not None:
            req = "UPDATE cw_cwattribute SET new_defaultval = %(val)s WHERE cw_eid = %(eid)s"
            args = {
                "val": dbh.binary_value(
                    convert_defaultval(cwattr, olddefault).getvalue()
                ),
                "eid": cwattr.eid,
            }
            sql(req, args, ask_confirm=False)

    sql("ALTER TABLE cw_cwattribute DROP COLUMN cw_defaultval")
    sql("ALTER TABLE cw_cwattribute RENAME COLUMN new_defaultval TO cw_defaultval")

    # Set object type to "Bytes" for CWAttribute's "defaultval" attribute
    rql(
        'SET X to_entity B WHERE X is CWAttribute, X from_entity Y, Y name "CWAttribute", '
        'X relation_type Z, Z name "defaultval", B name "Bytes", NOT X to_entity B'
    )

    oldrdef = schema["CWAttribute"].relation_definition("defaultval")
    import yams.buildobjs as ybo

    newrdef = ybo.RelationDefinition("CWAttribute", "defaultval", "Bytes")
    newrdef.eid = oldrdef.eid
    schema.add_relation_def(newrdef)
    schema.del_relation_def("CWAttribute", "defaultval", "String")

    commit()

    sync_schema_props_perms("defaultval")

    for rschema in schema.relations():
        if rschema.symmetric:
            subjects = set(repr(e.type) for e in rschema.subjects())
            objects = set(repr(e.type) for e in rschema.objects())
            assert subjects == objects
            martians = set(
                str(eid)
                for eid, in sql(
                    "SELECT eid_to FROM %s_relation, entities WHERE eid_to = eid AND type NOT IN (%s)"
                    % (rschema.type, ",".join(subjects))
                )
            )
            martians |= set(
                str(eid)
                for eid, in sql(
                    "SELECT eid_from FROM %s_relation, entities WHERE eid_from = eid AND type NOT IN (%s)"
                    % (rschema.type, ",".join(subjects))
                )
            )
            if martians:
                martians = ",".join(martians)
                print(f"deleting broken relations {rschema.type} for eids {martians}")
                sql(
                    "DELETE FROM %s_relation WHERE eid_from IN (%s) OR eid_to IN (%s)"
                    % (rschema.type, martians, martians)
                )
            with session.deny_all_hooks_but():
                rql(
                    "SET X %(r)s Y WHERE Y %(r)s X, NOT X %(r)s Y" % {"r": rschema.type}
                )
            commit()

    # multi columns unique constraints regeneration
    from cubicweb.server import schemaserial

    # syncschema hooks would try to remove indices but
    # 1) we already do that below
    # 2) the hook expects the CWUniqueTogetherConstraint.name attribute that hasn't
    #    yet been added
    with session.allow_all_hooks_but("syncschema"):
        rql("DELETE CWUniqueTogetherConstraint C")
    commit()
    add_attribute("CWUniqueTogetherConstraint", "name")

    # low-level wipe code for postgres, plain sql ...
    for (indexname,) in sql("select indexname from pg_indexes"):
        if indexname.startswith("unique_"):
            print("dropping index", indexname)
            sql(f"DROP INDEX {indexname}")
    commit()

    # recreate the constraints, hook will lead to low-level recreation
    for eschema in sorted(schema.entities()):
        if eschema._unique_together:
            print("recreate unique indexes for", eschema)
            rql_args = schemaserial.uniquetogether2rqls(eschema)
            for rql, args in rql_args:
                args["x"] = eschema.eid
                session.execute(rql, args)
    commit()

    # all attributes perms have to be refreshed ...
    for rschema in sorted(schema.relations()):
        if rschema.final:
            if rschema.type in fsschema:
                print("sync perms for", rschema.type)
                sync_schema_props_perms(
                    rschema.type, syncprops=False, ask_confirm=False, commit=False
                )
            else:
                print(f"WARNING: attribute {rschema.type} missing from fs schema")
    commit()

if applcubicwebversion < (3, 17, 0) and cubicwebversion >= (3, 17, 0):
    try:
        add_cube("sioc", update_database=False)
    except ConfigurationError:
        if not confirm(
            "In cubicweb 3.17 sioc views have been moved to the sioc "
            "cube, which is not installed.  Continue anyway?"
        ):
            raise
    try:
        add_cube("embed", update_database=False)
    except ConfigurationError:
        if not confirm(
            "In cubicweb 3.17 embedding views have been moved to the embed "
            "cube, which is not installed.  Continue anyway?"
        ):
            raise
    try:
        add_cube("geocoding", update_database=False)
    except ConfigurationError:
        if not confirm(
            "In cubicweb 3.17 geocoding views have been moved to the geocoding "
            "cube, which is not installed.  Continue anyway?"
        ):
            raise


if applcubicwebversion <= (3, 14, 0) and cubicwebversion >= (3, 14, 0):
    if "require_permission" in schema and "localperms" not in repo.config.cubes():
        from cubicweb import ExecutionError

        try:
            add_cube("localperms", update_database=False)
        except ConfigurationError:
            raise ExecutionError(
                "In cubicweb 3.14, CWPermission and related stuff "
                "has been moved to cube localperms. Install it first."
            )


if applcubicwebversion == (3, 6, 0) and cubicwebversion >= (3, 6, 0):
    CSTRMAP = dict(
        rql("Any T, X WHERE X is CWConstraintType, X name T", ask_confirm=False)
    )
    _add_relation_definition_no_perms("CWAttribute", "update_permission", "CWGroup")
    _add_relation_definition_no_perms(
        "CWAttribute", "update_permission", "RQLExpression"
    )
    rql("SET X update_permission Y WHERE X is CWAttribute, X add_permission Y")
    drop_relation_definition("CWAttribute", "delete_permission", "CWGroup")
    drop_relation_definition("CWAttribute", "delete_permission", "RQLExpression")

elif applcubicwebversion < (3, 6, 0) and cubicwebversion >= (3, 6, 0):
    CSTRMAP = dict(
        rql("Any T, X WHERE X is CWConstraintType, X name T", ask_confirm=False)
    )
    session.set_cnxset()
    permsdict = ss.deserialize_ertype_permissions(session)

    with session.allow_all_hooks_but("integrity"):
        for rschema in repo.schema.relations():
            rpermsdict = permsdict.get(rschema.eid, {})
            for rdef in rschema.relation_definitions.values():
                for action in rdef.ACTIONS:
                    actperms = []
                    for something in rpermsdict.get(
                        action == "update" and "add" or action, ()
                    ):
                        if isinstance(something, tuple):
                            actperms.append(rdef.rql_expression(*something))
                        else:  # group name
                            actperms.append(something)
                    rdef.set_action_permissions(action, actperms)
        for action in ("read", "add", "delete"):
            _add_relation_definition_no_perms(
                "CWRelation", f"{action}_permission", "CWGroup"
            )
            _add_relation_definition_no_perms(
                "CWRelation", f"{action}_permission", "RQLExpression"
            )
        for action in ("read", "update"):
            _add_relation_definition_no_perms(
                "CWAttribute", f"{action}_permission", "CWGroup"
            )
            _add_relation_definition_no_perms(
                "CWAttribute", f"{action}_permission", "RQLExpression"
            )
        for action in ("read", "add", "delete"):
            rql(
                "SET X %s_permission Y WHERE X is CWRelation, "
                "RT %s_permission Y, X relation_type RT, Y is CWGroup"
                % (action, action)
            )
            rql(
                "INSERT RQLExpression Y: Y exprtype YET, Y mainvars YMV, Y expression YEX, "
                "X %s_permission Y WHERE X is CWRelation, "
                "X relation_type RT, RT %s_permission Y2, Y2 exprtype YET, "
                "Y2 mainvars YMV, Y2 expression YEX" % (action, action)
            )
        rql(
            "SET X read_permission Y WHERE X is CWAttribute, "
            "RT read_permission Y, X relation_type RT, Y is CWGroup"
        )
        rql(
            "INSERT RQLExpression Y: Y exprtype YET, Y mainvars YMV, Y expression YEX, "
            "X read_permission Y WHERE X is CWAttribute, "
            "X relation_type RT, RT read_permission Y2, Y2 exprtype YET, "
            "Y2 mainvars YMV, Y2 expression YEX"
        )
        rql(
            "SET X update_permission Y WHERE X is CWAttribute, "
            "RT add_permission Y, X relation_type RT, Y is CWGroup"
        )
        rql(
            "INSERT RQLExpression Y: Y exprtype YET, Y mainvars YMV, Y expression YEX, "
            "X update_permission Y WHERE X is CWAttribute, "
            "X relation_type RT, RT add_permission Y2, Y2 exprtype YET, "
            "Y2 mainvars YMV, Y2 expression YEX"
        )
        for action in ("read", "add", "delete"):
            drop_relation_definition(
                "CWRType", f"{action}_permission", "CWGroup", commit=False
            )
            drop_relation_definition("CWRType", f"{action}_permission", "RQLExpression")
    sync_schema_props_perms(
        "read_permission", syncperms=False
    )  # fix read_permission cardinality

if (
    applcubicwebversion < (3, 9, 6)
    and cubicwebversion >= (3, 9, 6)
    and "CWUniqueTogetherConstraint" not in schema
):
    add_entity_type("CWUniqueTogetherConstraint")

if ("CWUniqueTogetherConstraint", "CWRType") not in schema[
    "relations"
].relation_definitions:
    add_relation_definition("CWUniqueTogetherConstraint", "relations", "CWRType")
    rql("SET C relations RT WHERE C relations RDEF, RDEF relation_type RT")
    commit()
    drop_relation_definition("CWUniqueTogetherConstraint", "relations", "CWAttribute")
    drop_relation_definition("CWUniqueTogetherConstraint", "relations", "CWRelation")

if applcubicwebversion < (3, 5, 0) and cubicwebversion >= (3, 5, 0):
    # check that migration is not doomed
    rset = rql(
        "Any X,Y WHERE X transition_of E, Y transition_of E, "
        "X name N, Y name N, NOT X identity Y",
        ask_confirm=False,
    )
    if rset:
        from logilab.common.shellutils import ASK

        if not ASK.confirm(
            "Migration will fail because of transitions with the same name. "
            "Continue anyway ?"
        ):
            import sys

            sys.exit(1)
    # proceed with migration
    add_entity_type("Workflow")
    add_entity_type("BaseTransition")
    add_entity_type("WorkflowTransition")
    add_entity_type("SubWorkflowExitPoint")
    # drop explicit 'State allowed_transition Transition' since it should be
    # infered due to yams inheritance.  However we've to disable the schema
    # sync hook first to avoid to destroy existing data...
    try:
        from cubicweb.hooks import syncschema

        repo.vreg.unregister(syncschema.AfterDelRelationTypeHook)
        try:
            drop_relation_definition("State", "allowed_transition", "Transition")
        finally:
            repo.vreg.register(syncschema.AfterDelRelationTypeHook)
    except ImportError:  # syncschema is in CW >= 3.6 only
        from cubicweb.server.schemahooks import after_del_relation_type

        repo.hm.unregister_hook(
            after_del_relation_type, "after_delete_relation", "relation_type"
        )
        try:
            drop_relation_definition("State", "allowed_transition", "Transition")
        finally:
            repo.hm.register_hook(
                after_del_relation_type, "after_delete_relation", "relation_type"
            )
    schema.rebuild_infered_relations()  # need to be explicitly called once everything is in place

    for et in rql(
        "DISTINCT Any ET,ETN WHERE S state_of ET, ET name ETN", ask_confirm=False
    ).entities():
        wf = add_workflow(f"default {et.name} workflow", et.name, ask_confirm=False)
        rql(
            "SET S state_of WF WHERE S state_of ET, ET eid %(et)s, WF eid %(wf)s",
            {"et": et.eid, "wf": wf.eid},
            "et",
            ask_confirm=False,
        )
        rql(
            "SET T transition_of WF WHERE T transition_of ET, ET eid %(et)s, WF eid %(wf)s",
            {"et": et.eid, "wf": wf.eid},
            "et",
            ask_confirm=False,
        )
        rql(
            "SET WF initial_state S WHERE ET initial_state S, ET eid %(et)s, WF eid %(wf)s",
            {"et": et.eid, "wf": wf.eid},
            "et",
            ask_confirm=False,
        )

    rql("DELETE TrInfo TI WHERE NOT TI from_state S")
    rql(
        "SET TI by_transition T WHERE TI from_state FS, TI to_state TS, "
        "FS allowed_transition T, T destination_state TS"
    )
    commit()

    drop_relation_definition("State", "state_of", "CWEType")
    drop_relation_definition("Transition", "transition_of", "CWEType")
    drop_relation_definition("CWEType", "initial_state", "State")

    sync_schema_props_perms()

if applcubicwebversion < (3, 2, 0) and cubicwebversion >= (3, 2, 0):
    add_cube("card", update_database=False)

if applcubicwebversion < (3, 20, 0) and cubicwebversion >= (3, 20, 0):
    ss._IGNORED_PROPS.append("formula")
    add_attribute("CWAttribute", "formula", commit=False)
    ss._IGNORED_PROPS.remove("formula")
    commit()
    add_entity_type("CWComputedRType")
    commit()

if schema["TZDatetime"].eid is None:
    add_entity_type("TZDatetime", auto=False)
if schema["TZTime"].eid is None:
    add_entity_type("TZTime", auto=False)


if applcubicwebversion < (3, 21, 1) and cubicwebversion >= (3, 21, 1):
    add_relation_definition("CWComputedRType", "read_permission", "CWGroup")
    add_relation_definition("CWComputedRType", "read_permission", "RQLExpression")


def sync_constraint_types():
    """Make sure the repository knows about all constraint types defined in the code"""
    from cubicweb.schema import CONSTRAINTS

    repo_constraints = set(
        row[0] for row in rql("Any N WHERE X is CWConstraintType, X name N")
    )

    for cstrtype in set(CONSTRAINTS) - repo_constraints:
        if cstrtype == "BoundConstraint":
            # was renamed to BoundaryConstraint, we don't need the old name
            continue
        rql("INSERT CWConstraintType X: X name %(name)s", {"name": cstrtype})

    commit()


sync_constraint_types()



# copyright 2003 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unit tests for module cubicweb.server.migractions"""

import os.path as osp

from cubicweb import Binary
from cubicweb.devtools import startpgcluster, stoppgcluster
from cubicweb.server.test.unittest_storage import StorageTCBase

from cubicweb.server.test_migractions.unittest_migractions import MigrationTC

HERE = osp.dirname(osp.abspath(__file__))
migrschema = None


def setUpModule():
    startpgcluster(__file__)


def tearDownModule(*args):
    global migrschema
    del migrschema
    stoppgcluster(__file__)


class MigrationStorageCommandsTC(StorageTCBase, MigrationTC):
    def test_change_bfss_path(self):
        with self.mh() as (cnx, mh):
            file1 = mh.cmd_create_entity(
                "File",
                data_name="foo.pdf",
                data=Binary(b"xxx"),
                data_format="text/plain",
            )
            mh.commit()
            current_dir = osp.dirname(self.fspath(cnx, file1))

            mh.update_bfss_path(current_dir, "loutre", commit=True)

            self.assertEqual("loutre", osp.dirname(self.fspath(cnx, file1)))


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2018 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

"""Tests for the entity edition"""

from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.server.edition import EditedEntity


class EditedEntityTC(CubicWebTC):
    """
    Test cases for EditedEntity
    """

    def test_clone_cache_reset(self):
        """
        Tests that when an EditedEntity is cloned the caches are reset in the cloned instance
        :return: Nothing
        """
        # Create an entity, create the EditedEntity and clone it
        with self.admin_access.cnx() as cnx:
            affaire = cnx.create_entity("Affaire", sujet="toto")
            ee = EditedEntity(affaire)
            ee.entity.cw_adapt_to("IWorkflowable")
            self.assertTrue(ee.entity._cw_related_cache)
            self.assertTrue(ee.entity._cw_adapters_cache)
            the_clone = ee.clone()
            self.assertFalse(the_clone.entity._cw_related_cache)
            self.assertFalse(the_clone.entity._cw_adapters_cache)
            cnx.rollback()
        # Check the attributes
        with self.admin_access.cnx() as cnx:
            # Assume a different connection set on the entity
            self.assertNotEqual(the_clone.entity._cw, cnx)
            # Use the new connection
            the_clone.entity._cw = cnx
            self.assertEqual("toto", the_clone.entity.sujet)


if __name__ == "__main__":
    import unittest

    unittest.main()



# pylint: disable=W0622,C0103
# copyright 2003-2025 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""cubicweb global packaging information for the cubicweb knowledge management
software
"""

from importlib.metadata import metadata

modname = distname = "cubicweb"
cubicweb_metadata = metadata(modname).json

version = cubicweb_metadata["version"]
numversion = [int(number) for number in version.split(".")]

description = cubicweb_metadata["description"]
author, author_email = cubicweb_metadata["author_email"].split()
web = cubicweb_metadata["project_url"][-1].split(", ")[-1]
license = cubicweb_metadata["license"]
classifiers = cubicweb_metadata["classifier"]



# copyright 2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# copyright 2014-2016 UNLISH S.A.S. (Montpellier, FRANCE), all rights reserved.
#
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""
CubicWeb AuthTkt authentication policy
--------------------------------------

When using the `cubicweb.pyramid.auth` module, which is the default in most
cases, you may have to configure the behaviour of these authentication
policies using standard's Pyramid configuration. You may want to configure in
your pyramid configuration file:

:Session Authentication:

    This is a `AuthTktAuthenticationPolicy`_ so you may overwrite default
    configuration values by adding configuration entries using the prefix
    ``cubicweb.auth.authtkt.session``. Default values are:

    ::

        cubicweb.auth.authtkt.session.hashalg = sha512
        cubicweb.auth.authtkt.session.cookie_name = auth_tkt
        cubicweb.auth.authtkt.session.timeout = 1200
        cubicweb.auth.authtkt.session.reissue_time = 120
        cubicweb.auth.authtkt.session.http_only = True
        cubicweb.auth.authtkt.session.secure = True


:Persistent Authentication:

    This is also a `AuthTktAuthenticationPolicy`_. It is used when persistent
    sessions are activated (typically when using the cubicweb-rememberme_
    cube). You may overwrite default configuration values by adding
    configuration entries using the prefix
    ``cubicweb.auth.authtkt.persistent``. Default values are:

    ::

        cubicweb.auth.authtkt.persistent.hashalg = sha512
        cubicweb.auth.authtkt.persistent.cookie_name = pauth_tkt
        cubicweb.auth.authtkt.persistent.max_age = 3600*24*30
        cubicweb.auth.authtkt.persistent.reissue_time = 3600*24
        cubicweb.auth.authtkt.persistent.http_only = True
        cubicweb.auth.authtkt.persistent.secure = True


.. Warning:: Legacy timeout values from the instance's
             ``all-in-one.conf`` are **not** used at all (``
             http-session-time`` and ``cleanup-session-time``)

.. _CubicWeb: https://www.cubicweb.org/
.. _`cubicweb-rememberme`: \
    https://forge.extranet.logilab.fr/cubicweb/cubes/rememberme
.. _AuthTktAuthenticationPolicy: \
    http://docs.pylonsproject.org/projects/pyramid/en/latest/api/authentication.html#pyramid.authentication.AuthTktAuthenticationPolicy


Secrets
~~~~~~~
There are a number of secrets to configure in ``pyramid.ini``. They
should be different one from each other, as explained in `Pyramid's
documentation`_.

For the record, regarding authentication:

:cubicweb.auth.authtkt.session.secret: This secret is used to encrypt
   the authentication cookie.

:cubicweb.auth.authtkt.persistent.secret: This secret is used to
   encrypt the persistent authentication cookie.
"""

import logging

from pyramid.authentication import AuthTktAuthenticationPolicy
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.interfaces import IAuthenticationPolicy
from pyramid.settings import asbool
from pyramid_multiauth import MultiAuthenticationPolicy
from zope.interface import implementer

from cubicweb.pyramid.core import get_principals
from cubicweb.pyramid.config import get_random_secret_key

log = logging.getLogger(__name__)


@implementer(IAuthenticationPolicy)
class UpdateLoginTimeAuthenticationPolicy:
    """An authentication policy that update the user last_login_time.

    The update is done in the 'remember' method, which is called by the login
    views login,

    Usually used via :func:`includeme`.
    """

    def authenticated_userid(self, request):
        pass

    def effective_principals(self, request):
        return ()

    def remember(self, request, principal, **kw):
        repo = request.registry["cubicweb.repository"]
        with repo.internal_cnx() as cnx:
            try:
                cnx.execute(
                    "SET U last_login_time NOW WHERE U eid %(user)s",
                    {"user": principal},
                )
                cnx.commit()
            except Exception:
                log.exception("Failed to update last_login_time")
        return ()

    def forget(self, request):
        return ()


class CWAuthTktAuthenticationPolicy(AuthTktAuthenticationPolicy):
    """
    An authentication policy that inhibate the call the 'remember' if a
    'persistent' argument is passed to it, and is equal to the value that
    was passed to the constructor.

    This allow to combine two policies with different settings and select them
    by just setting this argument.

    This also implement a ticket mechanism to invalidate a user cookie once it
    has logged out (to prevent a cookie reuse after a logout, pyramid doesn't
    provide this mechanism because it doesn't have any storage by default).

    This is done by storing a randomly generated token (that is an auth ticket)
    both in the user encrypted cookie and in its session. On logout, the ticket
    is marked as invalid in the session thus preventing any further use of this
    cookie.
    """

    def __init__(self, secret, persistent, defaults=None, prefix="", **settings):
        self.persistent = persistent
        unset = object()
        kw = {}
        if defaults is None:
            defaults = {}

        # load string settings
        for name in ("cookie_name", "path", "domain", "hashalg", "samesite"):
            value = settings.get(prefix + name, defaults.get(name, unset))
            if value is not unset:
                kw[name] = value

        # load boolean settings
        for name in (
            "secure",
            "include_ip",
            "http_only",
            "wild_domain",
            "parent_domain",
            "debug",
        ):
            value = settings.get(prefix + name, defaults.get(name, unset))
            if value is not unset:
                kw[name] = asbool(value)

        # load int settings
        for name in ("timeout", "reissue_time", "max_age"):
            value = settings.get(prefix + name, defaults.get(name, unset))
            if value is not unset and value is not None:
                kw[name] = int(value)

        super().__init__(secret, **kw)

    def _get_authentication_ticket_token(self, request):
        cookie_data = self.cookie.identify(request)
        if not cookie_data:
            return None

        authentication_ticket_token = [
            token for token in cookie_data["tokens"] if token.startswith("authticket_")
        ]

        if authentication_ticket_token:
            return authentication_ticket_token[0]
        else:
            return None

    def remember(self, request, userid, max_age=None, tokens=(), **kwargs):
        """
        Generate an auth ticket on login and stores it in the user session
        """

        # we are not handling a request for our kind of auth
        # for example it's a request for pauth_tkt but we are auth_tkt
        if "persistent" in kwargs and kwargs.pop("persistent") != self.persistent:
            return ()

        log.debug(f"generating an authentication ticket for user {userid}")

        # generate a new authentication ticket
        authentication_ticket_token = "authticket_" + get_random_secret_key()
        tokens += (authentication_ticket_token,)
        request.session["authentication_ticket"] = authentication_ticket_token

        return super().remember(
            request, userid, max_age=max_age, tokens=tokens, **kwargs
        )

    def unauthenticated_userid(self, request):
        """
        Validate that the current auth ticket, if it exists, is still valid
        """
        authentication_ticket_token = self._get_authentication_ticket_token(request)

        # auth hasn't finished yet
        if authentication_ticket_token is None:
            return None

        # no user stored in the cookie
        user_id = super().unauthenticated_userid(request)
        if not user_id:
            return None

        # likely when a new auth hasn't finished yet
        if authentication_ticket_token == request.session.get("authentication_ticket"):
            log.debug(
                "user %s authentication ticket is valid, validating user", user_id
            )
            return user_id
        else:
            log.debug(
                "could not find authentication ticket for user %s, preventing "
                "authentication (this could be a concurrent request during login or logout, not "
                "especially an attack",
                user_id,
            )
            return None

    def forget(self, request):
        """
        Invalidate the auth ticket if it's present
        """
        if "authentication_ticket" in request.session:
            log.debug("logout, deleting authentication ticket on request %s", request)
            # invalidate the ticket
            del request.session["authentication_ticket"]

        return super().forget(request)


def includeme(config):
    """Activate the CubicWeb AuthTkt authentication policy.

    It is automatically included by the configuration system, unless the
    following entry is added to the :ref:`pyramid_settings`:

    .. code-block:: ini

        cubicweb.pyramid.auth = no

    """
    settings = config.registry.settings

    policies = []

    if asbool(settings.get("cubicweb.auth.update_login_time", True)):
        policies.append(UpdateLoginTimeAuthenticationPolicy())

    if asbool(settings.get("cubicweb.auth.authtkt", True)):
        session_prefix = "cubicweb.auth.authtkt.session."
        persistent_prefix = "cubicweb.auth.authtkt.persistent."

        session_secret = settings.get(session_prefix + "secret", "notsosecret")
        persistent_secret = settings.get(persistent_prefix + "secret", "notsosecret")

        policies.append(
            CWAuthTktAuthenticationPolicy(
                session_secret,
                False,
                defaults={
                    "hashalg": "sha512",
                    "cookie_name": "auth_tkt",
                    "timeout": 1200,
                    "reissue_time": 120,
                    "http_only": True,
                    "secure": True,
                },
                prefix=session_prefix,
                **settings,
            )
        )

        policies.append(
            CWAuthTktAuthenticationPolicy(
                persistent_secret,
                True,
                defaults={
                    "hashalg": "sha512",
                    "cookie_name": "pauth_tkt",
                    "max_age": 3600 * 24 * 30,
                    "reissue_time": 3600 * 24,
                    "http_only": True,
                    "secure": True,
                },
                prefix=persistent_prefix,
                **settings,
            )
        )

    kw = {}
    if asbool(settings.get("cubicweb.auth.groups_principals", True)):
        kw["callback"] = get_principals

    authpolicy = MultiAuthenticationPolicy(policies, **kw)
    config.registry["cubicweb.authpolicy"] = authpolicy

    config.set_authentication_policy(authpolicy)
    config.set_authorization_policy(ACLAuthorizationPolicy())



# copyright 2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Tests for cubicweb.pyramid.config module."""

import os
from os import path
from tempfile import TemporaryDirectory
from unittest import TestCase
from unittest.mock import patch

from cubicweb.pyramid import config
from cubicweb.pyramid.pyramidctl import _generate_pyramid_ini_file


class PyramidConfigTC(TestCase):
    def test_get_random_secret_key(self):
        with patch("random.SystemRandom.choice", return_value="0") as patched_choice:
            secret = config.get_random_secret_key()
        self.assertEqual(patched_choice.call_count, 50)
        self.assertEqual(secret, "0" * 50)

    def test_write_pyramid_ini(self):
        with TemporaryDirectory() as instancedir:
            pyramid_ini_path = path.join(instancedir, "pyramid.ini")
            with patch(
                "random.SystemRandom.choice", return_value="0"
            ) as patched_choice:
                _generate_pyramid_ini_file(pyramid_ini_path)
            with open(path.join(instancedir, "pyramid.ini")) as f:
                lines = f.readlines()

        self.assertEqual(patched_choice.call_count, 50 * 3)

        secret = "0" * 50

        for option in (
            "cubicweb.session.secret",
            "cubicweb.auth.authtkt.persistent.secret",
            "cubicweb.auth.authtkt.session.secret",
        ):
            self.assertIn(f"{option} = {secret}\n", lines)

        for option in (
            "cubicweb.auth.authtkt.persistent.secure",
            "cubicweb.auth.authtkt.session.secure",
        ):
            self.assertIn(f"{option} = no\n", lines)

    def test_write_development_ini(self):
        with TemporaryDirectory() as instancedir:
            appid = "pyramid-instance"
            os.makedirs(path.join(instancedir, appid))
            os.environ["CW_INSTANCES_DIR"] = instancedir
            try:
                with patch(
                    "random.SystemRandom.choice", return_value="0"
                ) as patched_choice:
                    _generate_pyramid_ini_file(
                        os.path.join(instancedir, appid, "pyramid.ini")
                    )
            finally:
                os.environ.pop("CW_INSTANCES_DIR")
            with open(path.join(instancedir, appid, "pyramid.ini")) as f:
                lines = f.readlines()
        self.assertEqual(patched_choice.call_count, 50 * 3)
        secret = "0" * 50
        for option in (
            "cubicweb.session.secret",
            "cubicweb.auth.authtkt.persistent.secret",
            "cubicweb.auth.authtkt.session.secret",
        ):
            self.assertIn(f"{option} = {secret}\n", lines)
        self.assertIn("cubicweb.auth.authtkt.session.secure = no\n", lines)
        self.assertIn("cubicweb.auth.authtkt.persistent.secure = no\n", lines)


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.


import os
import tempfile
import unittest

from cubicweb.toolsutils import RQLExecuteMatcher, option_value_from_env, read_config


class RQLExecuteMatcherTests(unittest.TestCase):
    def matched_query(self, text):
        match = RQLExecuteMatcher.match(text)
        if match is None:
            return None
        return match["rql_query"]

    def test_unknown_function_dont_match(self):
        self.assertIsNone(self.matched_query("foo"))
        self.assertIsNone(self.matched_query("rql("))
        self.assertIsNone(self.matched_query('hell("")'))
        self.assertIsNone(self.matched_query("eval(\"rql('bla'"))

    def test_rql_other_parameters_dont_match(self):
        self.assertIsNone(self.matched_query('rql("Any X WHERE X eid %(x)s")'))
        self.assertIsNone(self.matched_query('rql("Any X WHERE X eid %(x)s", {'))
        self.assertIsNone(
            self.matched_query('session.execute("Any X WHERE X eid %(x)s")')
        )
        self.assertIsNone(
            self.matched_query('session.execute("Any X WHERE X eid %(x)s", {')
        )

    def test_rql_function_match(self):
        for func_expr in ("rql", "session.execute"):
            query = self.matched_query(f'{func_expr}("Any X WHERE X is ')
            self.assertEqual(query, "Any X WHERE X is ")

    def test_offseted_rql_function_match(self):
        """check indentation is allowed"""
        for func_expr in ("  rql", "  session.execute"):
            query = self.matched_query(f'{func_expr}("Any X WHERE X is ')
            self.assertEqual(query, "Any X WHERE X is ")


SOURCES_CONTENT = b"""
[admin]

# cubicweb manager account's login (this user will be created)
login=admin

# cubicweb manager account's password
password=admin

[system]

# database driver (postgres, sqlite)
db-driver=postgres

# database host
db-host=

# database port
db-port=
"""


class ToolsUtilsTC(unittest.TestCase):
    def test_option_value_from_env(self):
        os.environ["CW_DB_HOST"] = "here"
        try:
            self.assertEqual(option_value_from_env("db-host"), "here")
            self.assertEqual(option_value_from_env("db-host", "nothere"), "here")
            self.assertEqual(option_value_from_env("db-hots", "nothere"), "nothere")
        finally:
            del os.environ["CW_DB_HOST"]

    def test_read_config(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(SOURCES_CONTENT)
            f.seek(0)
            config = read_config(f.name)
        expected = {
            "admin": {
                "password": "admin",
                "login": "admin",
            },
            "system": {
                "db-port": None,
                "db-driver": "postgres",
                "db-host": None,
            },
        }
        self.assertEqual(config, expected)

    def test_read_config_env(self):
        os.environ["CW_DB_HOST"] = "here"
        try:
            with tempfile.NamedTemporaryFile() as f:
                f.write(SOURCES_CONTENT)
                f.seek(0)
                config = read_config(f.name)
        finally:
            del os.environ["CW_DB_HOST"]
        self.assertEqual(config["system"]["db-host"], "here")


if __name__ == "__main__":
    unittest.main()



# copyright 2003 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""some utilities to ease repository testing

This module contains functions to initialize a new repository.
"""
from contextlib import contextmanager
from pprint import pprint

from cubicweb import rqlrewrite
from cubicweb.entities.authobjs import user_session_cache_key
from cubicweb.server import rqlannotation
from cubicweb.server import set_debug, debugged
from cubicweb.server.querier import ExecutionPlan
from cubicweb.server.sources.rql2sql import remove_unused_solutions
from .testlib import RepoAccess, BaseTestCase


def tuplify(mylist):
    return [tuple(item) for item in mylist]


def snippet_key(a):
    # a[0] may be a dict or a key/value tuple
    return (sorted(dict(a[0]).items()), [e.expression for e in a[1]])


def check_plan(self, rql, expected, kwargs=None):
    with self.admin_access.cnx() as cnx:
        plan = self._prepare_plan(cnx, rql, kwargs)
        self.planner.build_plan(plan)
        try:
            self.assertEqual(
                len(plan.steps),
                len(expected),
                f"expected {len(expected)} steps, got {len(plan.steps)}",
            )
            # step order is important
            for i, step in enumerate(plan.steps):
                compare_steps(self, step.test_repr(), expected[i])
        except AssertionError:
            pprint([step.test_repr() for step in plan.steps])
            raise


def compare_steps(self, step, expected):
    try:
        self.assertEqual(
            step[0],
            expected[0],
            f"expected step type {expected[0]}, got {step[0]}",
        )
        if (
            len(step) > 2
            and isinstance(step[1], list)
            and isinstance(expected[1], list)
        ):
            queries, equeries = step[1], expected[1]
            self.assertEqual(
                len(queries),
                len(equeries),
                f"expected {len(equeries)} queries, got {len(queries)}",
            )
            for i, (rql, sol) in enumerate(queries):
                self.assertEqual(rql, equeries[i][0])
                self.assertEqual(
                    sorted(sorted(x.items()) for x in sol),
                    sorted(sorted(x.items()) for x in equeries[i][1]),
                )
            idx = 2
        else:
            idx = 1
        self.assertEqual(
            step[idx:-1],
            expected[idx:-1],
            f"expected step characteristic \n{expected[1:-1]}\n, got\n{step[1:-1]}",
        )
        self.assertEqual(
            len(step[-1]),
            len(expected[-1]),
            f"got {len(step[-1])} child steps, expected {len(expected[-1])}",
        )
    except AssertionError:
        print("error on step ", end=" ")
        pprint(step[:-1])
        raise
    children = step[-1]
    if step[0] in ("UnionFetchStep", "UnionStep"):
        # sort children
        children = sorted(children)
        expectedchildren = sorted(expected[-1])
    else:
        expectedchildren = expected[-1]
    for i, substep in enumerate(children):
        compare_steps(self, substep, expectedchildren[i])


class DumbOrderedDict(list):
    def __iter__(self):
        return self.iterkeys()

    def __contains__(self, key):
        return key in self.iterkeys()

    def __getitem__(self, key):
        for key_, value in list.__iter__(self):
            if key == key_:
                return value
        raise KeyError(key)

    def iterkeys(self):
        return (x for x, y in list.__iter__(self))

    def iteritems(self):
        return (x for x in list.__iter__(self))

    def items(self):
        return [x for x in list.__iter__(self)]


def schema_eids_idx(schema):
    """return a dictionary mapping schema types to their eids so we can reread
    it from the fs instead of the db (too costly) between tests
    """
    schema_eids = {}
    for x in schema.entities():
        schema_eids[x] = x.eid
    for x in schema.relations():
        schema_eids[x] = x.eid
        for rdef in x.relation_definitions.values():
            schema_eids[(rdef.subject, rdef.relation_type, rdef.object)] = rdef.eid
    return schema_eids


def restore_schema_eids_idx(schema, schema_eids):
    """rebuild schema eid index"""
    for x in schema.entities():
        x.eid = schema_eids[x]
        schema._eid_index[x.eid] = x
    for x in schema.relations():
        x.eid = schema_eids[x]
        schema._eid_index[x.eid] = x
        for rdef in x.relation_definitions.values():
            rdef.eid = schema_eids[(rdef.subject, rdef.rtype, rdef.object)]
            schema._eid_index[rdef.eid] = rdef


class BaseQuerierTC(BaseTestCase):
    repo = None  # set this in concrete class

    def setUp(self):
        self.o = self.repo.querier
        self.admin_access = RepoAccess(self.repo, "admin")
        self.ueid = self.admin_access._user.eid
        assert self.ueid != -1
        self.repo._type_cache = {}  # clear cache
        do_monkey_patch()
        self._dumb_sessions = []

    def tearDown(self):
        undo_monkey_patch()
        assert self.admin_access._user.eid != -1

    def set_debug(self, debug):
        set_debug(debug)

    def debugged(self, debug):
        return debugged(debug)

    def _rqlhelper(self):
        rqlhelper = self.repo.vreg.rqlhelper
        # reset uid_func so it don't try to get type from eids
        rqlhelper._analyser.uid_func = None
        rqlhelper._analyser.uid_func_mapping = {}
        return rqlhelper

    def _prepare_plan(self, cnx, rql, kwargs=None, simplify=True):
        rqlhelper = self._rqlhelper()
        rqlst = rqlhelper.parse(rql)
        rqlhelper.compute_solutions(rqlst, kwargs=kwargs)
        if simplify:
            rqlhelper.simplify(rqlst)
        for select in rqlst.children:
            select.solutions.sort(key=lambda x: list(x.items()))
        return self.o.plan_factory(rqlst, kwargs, cnx)

    def _prepare(self, cnx, rql, kwargs=None):
        plan = self._prepare_plan(cnx, rql, kwargs, simplify=False)
        plan.preprocess(plan.rqlst)
        rqlst = plan.rqlst.children[0]
        rqlst.solutions = remove_unused_solutions(
            rqlst, rqlst.solutions, self.repo.schema
        )[0]
        return rqlst

    @contextmanager
    def user_groups_session(self, *groups):
        """lightweight session using the current user with hi-jacked groups"""
        # use cnx.user.eid to get correct owned_by relation, unless explicit eid
        with self.admin_access.cnx() as cnx:
            user_eid = cnx.user.eid
            cnx.user._cw.transaction_data[
                user_session_cache_key(user_eid, "groups")
            ] = set(groups)
            yield cnx

    def qexecute(self, rql, args=None, build_descr=True):
        with self.admin_access.cnx() as cnx:
            try:
                return self.o.execute(cnx, rql, args, build_descr)
            finally:
                if rql.startswith(("INSERT", "DELETE", "SET")):
                    cnx.commit()


class BasePlannerTC(BaseQuerierTC):
    def setup(self):
        # XXX source_defs
        self.o = self.repo.querier
        self.schema = self.o.schema
        self.system = self.repo.system_source
        do_monkey_patch()
        self.repo.vreg.rqlhelper.backend = "postgres"  # so FTIRANK is considered

    def tearDown(self):
        undo_monkey_patch()

    def _prepare_plan(self, cnx, rql, kwargs=None):
        rqlst = self.repo.vreg.rqlhelper.parse(rql, annotate=True)
        self.repo.vreg.compute_var_types(cnx, rqlst, kwargs)
        if rqlst.TYPE == "select":
            self.repo.vreg.rqlhelper.annotate(rqlst)
            for select in rqlst.children:
                select.solutions.sort(key=lambda x: list(x.items()))
        else:
            rqlst.solutions.sort(key=lambda x: list(x.items()))
        return self.o.plan_factory(rqlst, kwargs, cnx)


# monkey patch some methods to get predictable results #######################

_orig_iter_relations = rqlrewrite.iter_relations
_orig_insert_snippets = rqlrewrite.RQLRewriter.insert_snippets
_orig_build_variantes = rqlrewrite.RQLRewriter.build_variantes


def _insert_snippets(self, snippets, varexistsmap=None):
    _orig_insert_snippets(self, sorted(snippets, key=snippet_key), varexistsmap)


def _build_variantes(self, newsolutions):
    variantes = _orig_build_variantes(self, newsolutions)
    sortedvariantes = []
    for variante in variantes:
        orderedkeys = sorted((k[1], k[2], v) for k, v in variante.items())
        variante = DumbOrderedDict(
            sorted(variante.items(), key=lambda a: (a[0][1], a[0][2], a[1]))
        )
        sortedvariantes.append((orderedkeys, variante))
    return [v for ok, v in sorted(sortedvariantes)]


_orig_check_permissions = ExecutionPlan._check_permissions


def _check_permissions(*args, **kwargs):
    res, restricted = _orig_check_permissions(*args, **kwargs)
    res = DumbOrderedDict(
        sorted(res.items(), key=lambda x: [list(y.items()) for y in x[1]])
    )
    return res, restricted


_orig_select_principal = rqlannotation._select_principal


def _select_principal(scope, relations):
    def sort_key(something):
        try:
            return something.r_type
        except AttributeError:
            return (something[0].r_type, something[1])

    return _orig_select_principal(
        scope, relations, _sort=lambda rels: sorted(rels, key=sort_key)
    )


def _ordered_iter_relations(stinfo):
    return sorted(_orig_iter_relations(stinfo), key=lambda x: x.r_type)


def do_monkey_patch():
    rqlrewrite.iter_relations = _ordered_iter_relations
    rqlrewrite.RQLRewriter.insert_snippets = _insert_snippets
    rqlrewrite.RQLRewriter.build_variantes = _build_variantes
    ExecutionPlan._check_permissions = _check_permissions
    ExecutionPlan.tablesinorder = None


def undo_monkey_patch():
    rqlrewrite.iter_relations = _orig_iter_relations
    rqlrewrite.RQLRewriter.insert_snippets = _orig_insert_snippets
    rqlrewrite.RQLRewriter.build_variantes = _orig_build_variantes
    ExecutionPlan._check_permissions = _orig_check_permissions



# copyright 2003-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unit tests for i18n messages generator"""

import os.path as osp
import sys
from contextlib import contextmanager
from io import StringIO
from unittest import main
from unittest.mock import patch

from cubicweb.devtools import devctl
from cubicweb.devtools.testlib import BaseTestCase

DATADIR = osp.join(osp.abspath(osp.dirname(__file__)), "data")


TESTCUBE_DIR = osp.join(DATADIR, "libpython", "cubicweb_i18ntestcube")


class CustomMessageExtractor(devctl.I18nCubeMessageExtractor):
    blacklist = devctl.I18nCubeMessageExtractor.blacklist | {"excludeme"}


@contextmanager
def capture_stdout():
    stream = StringIO()
    sys.stdout = stream
    yield stream
    stream.seek(0)
    sys.stdout = sys.__stdout__


class I18nCollectorTest(BaseTestCase):
    def test_i18ncube_py_collection(self):
        extractor = CustomMessageExtractor(DATADIR, TESTCUBE_DIR)
        collected = extractor.collect_py()
        expected = [
            osp.join(TESTCUBE_DIR, path)
            for path in ("__init__.py", "__pkginfo__.py", "schema.py")
        ]
        self.assertCountEqual(expected, collected)

    def test_i18ncube_js_collection(self):
        extractor = CustomMessageExtractor(DATADIR, TESTCUBE_DIR)
        collected = extractor.collect_js()
        self.assertCountEqual([], collected, [])
        extractor.blacklist = ()  # don't ignore anything
        collected = extractor.collect_js()
        expected = [osp.join(TESTCUBE_DIR, "node_modules/cubes.somefile.js")]
        self.assertCountEqual(expected, collected)

    class FakeMessageExtractor(devctl.I18nCubeMessageExtractor):
        """Fake message extractor that generates no pot file."""

        def generate_pot_file(self):
            return None

    @patch("pkg_resources.load_entry_point", return_value=FakeMessageExtractor)
    def test_cube_custom_extractor(self, mock_load_entry_point):
        distname = "cubicweb_i18ntestcube"  # same for new and legacy layout
        cubedir = osp.join(DATADIR, "libpython", "cubicweb_i18ntestcube")
        with capture_stdout() as stream:
            devctl.update_cube_catalogs(cubedir)
        self.assertIn("no message catalog for cube i18ntestcube", stream.read())
        mock_load_entry_point.assert_called_once_with(
            distname, "cubicweb.i18ncube", "i18ntestcube"
        )
        mock_load_entry_point.reset_mock()


if __name__ == "__main__":
    main()



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from yams.buildobjs import (
    EntityType,
    RelationType,
    RelationDefinition,
    ComputedRelation,
    SubjectRelation,
    RichString,
    String,
    Int,
    Float,
    Boolean,
    Datetime,
    TZDatetime,
    Bytes,
)
from yams.constraints import SizeConstraint

from cubicweb import _
from cubicweb.schema import (
    WorkflowableEntityType,
    RQLUniqueConstraint,
    RQLVocabularyConstraint,
    ERQLExpression,
    RRQLExpression,
)
from cubicweb.schemas.base import CWUser


class Affaire(WorkflowableEntityType):
    __permissions__ = {
        "read": (
            "managers",
            ERQLExpression("X owned_by U"),
            ERQLExpression("X concerne S?, S owned_by U"),
        ),
        "add": ("managers", ERQLExpression("X concerne S, S owned_by U")),
        "update": (
            "managers",
            "owners",
            ERQLExpression('X in_state S, S name in ("pitetre", "en cours")'),
        ),
        "delete": ("managers", "owners", ERQLExpression("X concerne S, S owned_by U")),
    }

    ref = String(
        fulltextindexed=True, indexed=True, constraints=[SizeConstraint(max=16)]
    )
    sujet = String(fulltextindexed=True, constraints=[SizeConstraint(max=256)])
    descr = RichString(fulltextindexed=True, description=_("more detailed description"))

    duration = Int()
    invoiced = Float()
    opt_attr = Bytes()

    depends_on = SubjectRelation("Affaire")
    require_permission = SubjectRelation("CWPermission")
    concerne = SubjectRelation(("Societe", "Note"))
    todo_by = SubjectRelation("Personne", cardinality="?*")
    documented_by = SubjectRelation("Card")


class Societe(EntityType):
    __unique_together__ = [("nom", "type", "cp")]
    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "update": ("managers", "owners", ERQLExpression("U login L, X nom L")),
        "delete": ("managers", "owners", ERQLExpression("U login L, X nom L")),
        "add": (
            "managers",
            "users",
        ),
    }

    nom = String(maxsize=64, fulltextindexed=True)
    web = String(maxsize=128)
    type = String(maxsize=128)  # attribute in common with Note
    tel = Int()
    fax = Int()
    rncs = String(maxsize=128)
    ad1 = String(maxsize=128)
    ad2 = String(maxsize=128)
    ad3 = String(maxsize=128)
    cp = String(maxsize=12)
    ville = String(maxsize=32)


class Division(Societe):
    __specializes_schema__ = True


class SubDivision(Division):
    __specializes_schema__ = True


class travaille_subdivision(RelationDefinition):
    subject = "Personne"
    object = "SubDivision"


next(CWUser.get_relations("login")).fulltextindexed = True


class Note(WorkflowableEntityType):
    date = String(maxsize=10)
    type = String(vocabulary=["todo", "a", "b", "T", "lalala"])
    para = String(
        maxsize=512,
        __permissions__={
            "add": ("managers", ERQLExpression('X in_state S, S name "todo"')),
            "read": ("managers", "users", "guests"),
            "update": ("managers", ERQLExpression('X in_state S, S name "todo"')),
        },
    )
    something = String(
        maxsize=1,
        __permissions__={
            "read": ("managers", "users", "guests"),
            "add": (ERQLExpression("NOT X para NULL"),),
            "update": ("managers", "owners"),
        },
    )
    migrated_from = SubjectRelation("Note")
    attachment = SubjectRelation("File")
    inline1 = SubjectRelation(
        "Affaire",
        inlined=True,
        cardinality="?*",
        constraints=[
            RQLUniqueConstraint(
                "S type T, S inline1 A1, A1 todo_by C, "
                "Y type T, Y inline1 A2, A2 todo_by C",
                "S,Y",
            )
        ],
    )
    todo_by = SubjectRelation("CWUser")


class Frozable(EntityType):
    __permissions__ = {
        "read": ("managers", "users"),
        "add": ("managers", "users"),
        "update": (
            "managers",
            ERQLExpression("X frozen False"),
        ),
        "delete": (
            "managers",
            ERQLExpression("X frozen False"),
        ),
    }
    name = String()
    frozen = Boolean(
        default=False,
        __permissions__={
            "read": ("managers", "users"),
            "add": ("managers", "users"),
            "update": ("managers", "owners"),
        },
    )


class Personne(EntityType):
    __permissions__ = {
        "read": ("managers", "users", "guests"),  # 'guests' will be removed
        "add": ("managers", "users"),
        "update": ("managers", "owners"),
        "delete": ("managers", "owners"),
    }
    __unique_together__ = [("nom", "prenom", "inline2")]
    nom = String(fulltextindexed=True, required=True, maxsize=64)
    prenom = String(fulltextindexed=True, maxsize=64)
    sexe = String(maxsize=1, default="M", fulltextindexed=True)
    promo = String(vocabulary=("bon", "pasbon"))
    titre = String(fulltextindexed=True, maxsize=128)
    adel = String(maxsize=128)
    ass = String(maxsize=128)
    web = String(maxsize=128)
    tel = Int()
    fax = Int()
    datenaiss = Datetime()
    tzdatenaiss = TZDatetime()
    test = Boolean(
        __permissions__={
            "read": ("managers", "users", "guests"),
            "add": ("managers",),
            "update": ("managers",),
        }
    )
    description = String()
    firstname = String(fulltextindexed=True, maxsize=64)

    concerne = SubjectRelation("Affaire")
    connait = SubjectRelation("Personne")
    inline2 = SubjectRelation("Affaire", inlined=True, cardinality="?*")


class Old(EntityType):
    name = String(
        __permissions__={
            "read": ("managers", "users", "guests"),
            "add": ("managers", "users", "guests"),
            "update": (),
        }
    )


class Email(EntityType):
    subject = String(fulltextindexed=True)
    messageid = String(required=True, indexed=True, unique=True)
    sender = SubjectRelation("EmailAddress", cardinality="?*")
    recipients = SubjectRelation("EmailAddress")
    attachment = SubjectRelation("File")


class EmailPart(EntityType):
    pass


class EmailThread(EntityType):
    see_also = SubjectRelation("EmailThread")


class connait(RelationType):
    symmetric = True


class concerne(RelationType):
    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": ("managers", RRQLExpression("U has_update_permission S")),
        "delete": ("managers", RRQLExpression("O owned_by U")),
    }


class travaille(RelationDefinition):
    __permissions__ = {
        "read": ("managers", "users", "guests"),
        "add": ("managers", RRQLExpression("U has_update_permission S")),
        "delete": ("managers", RRQLExpression("O owned_by U")),
    }
    subject = "Personne"
    object = "Societe"
    constraints = [
        RQLVocabularyConstraint("S owned_by U"),
        RQLVocabularyConstraint("S created_by U"),
    ]


class comments(RelationDefinition):
    subject = "Comment"
    object = "Personne"


class fiche(RelationDefinition):
    inlined = True
    subject = "Personne"
    object = "Card"
    cardinality = "??"


class multisource_inlined_rel(RelationDefinition):
    inlined = True
    cardinality = "?*"
    subject = ("Card", "Note")
    object = ("Affaire", "Note")


class see_also_1(RelationDefinition):
    name = "see_also"
    subject = object = "Folder"


class see_also_2(RelationDefinition):
    name = "see_also"
    subject = ("Bookmark", "Note")
    object = ("Bookmark", "Note")


class evaluee(RelationDefinition):
    subject = ("Personne", "CWUser", "Societe")
    object = "Note"
    constraints = [
        RQLVocabularyConstraint("S created_by U"),
        RQLVocabularyConstraint("S owned_by U"),
    ]


class ecrit_par(RelationType):
    inlined = True


class ecrit_par_1(RelationDefinition):
    name = "ecrit_par"
    subject = "Note"
    object = "Personne"
    cardinality = "?*"


class ecrit_par_2(RelationDefinition):
    name = "ecrit_par"
    subject = "Note"
    object = "CWUser"
    cardinality = "?*"


class copain(RelationDefinition):
    subject = object = "CWUser"


class tags(RelationDefinition):
    subject = "Tag"
    object = ("CWUser", "CWGroup", "State", "Note", "Card", "Affaire")


class Folder(EntityType):
    """folders are used to classify entities. They may be defined as a tree."""

    name = String(required=True, indexed=True, internationalizable=True, maxsize=64)
    description = RichString(fulltextindexed=True)
    filed_under = SubjectRelation("Folder", description=_("parent folder"))


class filed_under(RelationDefinition):
    subject = ("Note", "Affaire")
    object = "Folder"


class require_permission(RelationDefinition):
    subject = ("Card", "Note", "Personne")
    object = "CWPermission"


class require_state(RelationDefinition):
    subject = "CWPermission"
    object = "State"


class personne_composite(RelationDefinition):
    subject = "Personne"
    object = "Personne"
    composite = "subject"


class personne_inlined(RelationDefinition):
    subject = "Personne"
    object = "Personne"
    cardinality = "?*"
    inlined = True


class login_user(RelationDefinition):
    subject = "Personne"
    object = "CWUser"
    cardinality = "??"


class ambiguous_inlined(RelationDefinition):
    subject = ("Affaire", "Note")
    object = "CWUser"
    inlined = True
    cardinality = "?*"


class user_login(ComputedRelation):
    rule = "O login_user S"



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""functional tests for integrity hooks"""

from yams import ValidationError
from cubicweb.devtools.testlib import CubicWebTC


class CoreHooksTC(CubicWebTC):
    def test_delete_internal_entities(self):
        with self.admin_access.repo_cnx() as cnx:
            self.assertRaises(
                ValidationError, cnx.execute, 'DELETE CWEType X WHERE X name "CWEType"'
            )
            cnx.rollback()
            self.assertRaises(
                ValidationError,
                cnx.execute,
                'DELETE CWRType X WHERE X name "relation_type"',
            )
            cnx.rollback()
            self.assertRaises(
                ValidationError, cnx.execute, 'DELETE CWGroup X WHERE X name "owners"'
            )

    def test_delete_required_relations_subject(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.execute(
                'INSERT CWUser X: X login "toto", X upassword "hop", X in_group Y '
                'WHERE Y name "users"'
            )
            cnx.commit()
            cnx.execute('DELETE X in_group Y WHERE X login "toto", Y name "users"')
            self.assertRaises(ValidationError, cnx.commit)
            cnx.rollback()
            cnx.execute('DELETE X in_group Y WHERE X login "toto"')
            cnx.execute('SET X in_group Y WHERE X login "toto", Y name "guests"')
            cnx.commit()

    def test_static_vocabulary_check(self):
        with self.admin_access.repo_cnx() as cnx:
            self.assertRaises(
                ValidationError,
                cnx.execute,
                'SET X composite "whatever" WHERE X from_entity FE, FE name "CWUser", '
                'X relation_type RT, RT name "in_group"',
            )

    def test_missing_required_relations_subject_inline(self):
        with self.admin_access.repo_cnx() as cnx:
            # missing in_group relation
            cnx.execute('INSERT CWUser X: X login "toto", X upassword "hop"')
            self.assertRaises(ValidationError, cnx.commit)

    def test_composite_1(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.execute(
                'INSERT EmailAddress X: X address "toto@logilab.fr", X alias "hop"'
            )
            cnx.execute(
                'INSERT EmailPart X: X content_format "text/plain", X ordernum 1, '
                'X content "this is a test"'
            )
            cnx.execute(
                'INSERT Email X: X messageid "<1234>", X subject "test", X sender Y, '
                "X recipients Y, X parts P "
                "WHERE Y is EmailAddress, P is EmailPart"
            )
            self.assertTrue(cnx.execute("Email X WHERE X sender Y"))
            cnx.commit()
            cnx.execute("DELETE Email X")
            rset = cnx.execute("Any X WHERE X is EmailPart")
            self.assertEqual(len(rset), 0)
            cnx.commit()
            rset = cnx.execute("Any X WHERE X is EmailPart")
            self.assertEqual(len(rset), 0)

    def test_composite_2(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.execute(
                'INSERT EmailAddress X: X address "toto@logilab.fr", X alias "hop"'
            )
            cnx.execute(
                'INSERT EmailPart X: X content_format "text/plain", X ordernum 1, '
                'X content "this is a test"'
            )
            cnx.execute(
                'INSERT Email X: X messageid "<1234>", X subject "test", X sender Y, '
                "X recipients Y, X parts P "
                "WHERE Y is EmailAddress, P is EmailPart"
            )
            cnx.commit()
            cnx.execute("DELETE Email X")
            cnx.execute("DELETE EmailPart X")
            cnx.commit()
            rset = cnx.execute("Any X WHERE X is EmailPart")
            self.assertEqual(len(rset), 0)

    def test_composite_redirection(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.execute(
                'INSERT EmailAddress X: X address "toto@logilab.fr", X alias "hop"'
            )
            cnx.execute(
                'INSERT EmailPart X: X content_format "text/plain", X ordernum 1, '
                'X content "this is a test"'
            )
            cnx.execute(
                'INSERT Email X: X messageid "<1234>", X subject "test", X sender Y, '
                "X recipients Y, X parts P "
                "WHERE Y is EmailAddress, P is EmailPart"
            )
            cnx.execute(
                'INSERT Email X: X messageid "<2345>", X subject "test2", X sender Y, '
                "X recipients Y "
                "WHERE Y is EmailAddress"
            )
            cnx.commit()
            cnx.execute('DELETE X parts Y WHERE X messageid "<1234>"')
            cnx.execute('SET X parts Y WHERE X messageid "<2345>"')
            cnx.commit()
            rset = cnx.execute("Any X WHERE X is EmailPart")
            self.assertEqual(len(rset), 1)
            self.assertEqual(rset.get_entity(0, 0).reverse_parts[0].messageid, "<2345>")

    def test_composite_object_relation_deletion(self):
        with self.admin_access.repo_cnx() as cnx:
            root = cnx.create_entity("Folder", name="root")
            a = cnx.create_entity("Folder", name="a", parent=root)
            cnx.create_entity("Folder", name="b", parent=a)
            cnx.create_entity("Folder", name="c", parent=root)
            cnx.commit()
            cnx.execute('DELETE Folder F WHERE F name "a"')
            cnx.execute('DELETE F parent R WHERE R name "root"')
            cnx.commit()
            self.assertEqual(
                [["root"], ["c"]],
                cnx.execute("Any NF WHERE F is Folder, F name NF").rows,
            )
            self.assertEqual(
                [], cnx.execute("Any NF,NP WHERE F parent P, F name NF, P name NP").rows
            )

    def test_composite_subject_relation_deletion(self):
        with self.admin_access.repo_cnx() as cnx:
            root = cnx.create_entity("Folder", name="root")
            a = cnx.create_entity("Folder", name="a")
            b = cnx.create_entity("Folder", name="b")
            c = cnx.create_entity("Folder", name="c")
            root.cw_set(children=(a, c))
            a.cw_set(children=b)
            cnx.commit()
            cnx.execute('DELETE Folder F WHERE F name "a"')
            cnx.execute('DELETE R children F WHERE R name "root"')
            cnx.commit()
            self.assertEqual(
                [["root"], ["c"]],
                cnx.execute("Any NF WHERE F is Folder, F name NF").rows,
            )
            self.assertEqual(
                [], cnx.execute("Any NF,NP WHERE F parent P, F name NF, P name NP").rows
            )

    def test_unsatisfied_constraints(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.execute('SET U in_group G WHERE G name "owners", U login "admin"')[0][0]
            with self.assertRaises(ValidationError) as cm:
                cnx.commit()
        self.assertEqual(
            cm.exception.errors,
            {"in_group-object": 'RQLConstraint NOT O name "owners" failed'},
        )

    def test_unique_constraint(self):
        with self.admin_access.repo_cnx() as cnx:
            entity = cnx.create_entity("CWGroup", name="trout")
            cnx.commit()
            self.assertRaises(
                ValidationError, cnx.create_entity, "CWGroup", name="trout"
            )
            cnx.rollback()
            cnx.execute('SET X name "trout" WHERE X eid %(x)s', {"x": entity.eid})
            cnx.commit()


if __name__ == "__main__":
    from logilab.common.testlib import unittest_main

    unittest_main()
